import pytest
from ape import Contract
from ape.contracts import ContractInstance
from ape.exceptions import ContractLogicError, OutOfGasError


@pytest.fixture(scope="module", autouse=True)
def connection(provider):
    yield


def test_is_token(contract, tokens):
    assert not tokens.is_token(contract.address)


def test_deploy(project):
    assert project.MyContract, "Unable to access contract when needing to compile"

    contract = project.MyContract
    assert contract, "Unable to access contract when not needing to compile"

    deployment = contract.deploy()
    assert deployment


def test_declare_then_deploy(account, chain, project, provider, factory_contract_container):
    # Declare contract type. The result should contain a 'class_hash'.
    declaration = provider.declare(project.MyContract)
    assert declaration.class_hash

    # Deploy a ContractInstance from the declaration.
    contract = declaration.deploy()
    assert isinstance(contract, ContractInstance)

    # Ensure can interact with deployed contract from declaration.
    contract.initialize(sender=account)
    balance_pre_call = contract.get_balance(account)
    contract.increase_balance(account, 9, sender=account)
    assert contract.get_balance(account) == balance_pre_call + 9

    # Ensure can use class_hash in factory contract
    factory = factory_contract_container.deploy(declaration.class_hash)
    receipt = factory.create_my_contract(sender=account)
    logs = list(receipt.decode_logs(factory.contract_deployed))
    new_contract_address = provider.starknet.decode_address(logs[0]["contract_address"])

    # # Ensure can interact with deployed contract from 'class_hash'.
    new_contract_instance = Contract(new_contract_address, contract_type=contract.contract_type)
    assert new_contract_instance
    new_contract_instance.initialize(sender=account)
    balance_pre_call = new_contract_instance.get_balance(account)
    new_contract_instance.increase_balance(account, 9, sender=account)
    assert new_contract_instance.get_balance(account) == balance_pre_call + 9


def test_get_caller_address(contract, account, provider):
    expected = provider.starknet.encode_address(account.address)
    assert contract.get_caller(sender=account).return_value == expected


def test_validate_signature_on_chain(contract, account, initial_balance):
    # NOTE: This test validates the account signature but the transaction
    # is not directly sent from the account.
    increase_amount = 42 * 2**152

    signature = account.sign_message(increase_amount)
    contract.increase_balance_signed(
        account.public_key, account.address, increase_amount, signature
    )

    actual = contract.get_balance(account)
    expected = initial_balance + increase_amount
    assert actual == expected


def test_transact_from_account(contract, account, initial_balance):
    increase_amount = 123456
    receipt = contract.increase_balance(account, increase_amount, sender=account)
    actual_from_receipt = receipt.return_value
    actual_from_call = contract.get_balance(account)
    expected = initial_balance + increase_amount
    assert actual_from_receipt == actual_from_call == expected


def test_contracts_as_arguments(contract, account):
    initial_balance = contract.get_balance(contract)
    increase_amount = 123456
    receipt = contract.increase_balance(contract, increase_amount, sender=account)
    actual_from_receipt = receipt.return_value
    actual_from_call = contract.get_balance(contract)
    expected = initial_balance + increase_amount
    assert actual_from_receipt == actual_from_call == expected


def test_unsigned_contract_transaction(contract, account, initial_balance):
    increase_amount = 123456
    receipt = contract.increase_balance(account.address, increase_amount)

    actual_from_receipt = receipt.return_value
    actual_from_call = contract.get_balance(account.address)
    expected = initial_balance + increase_amount
    assert actual_from_receipt == actual_from_call == expected


def test_decode_logs(contract, account, ecosystem):
    increase_amount = 9933
    receipt = contract.increase_balance(account.address, increase_amount, sender=account)
    logs = list(receipt.decode_logs(contract.balance_increased))
    assert len(logs) == 1
    assert logs[0].amount == increase_amount

    from_address = receipt.logs[0]["from_address"]
    log_sender_address = ecosystem.decode_address(from_address)
    assert log_sender_address == contract.address


def test_revert_message(contract):
    reason = "Already initialized"
    with pytest.raises(ContractLogicError, match=reason):
        # Already initialized from fixture
        contract.initialize()


def test_revert_no_message(contract, account):
    contract.reset()
    reason = "An ASSERT_EQ instruction failed.*"
    with pytest.raises(ContractLogicError, match=reason):
        contract.increase_balance(account.address, 123)

    # Re-initialize (re-store state)
    contract.initialize()


def test_unable_to_afford_transaction(contract, account, provider):
    with pytest.raises(OutOfGasError):
        contract.increase_balance(account.address, 1, sender=account, max_fee=1)


def test_array_inputs(contract):
    # This test makes sure we can pass python lists as arguments
    # to Cairo methods that accept arrays.
    # NOTE: Due to a limitation in ape, we have to include the array length argument.
    contract.store_sum(3, [1, 2, 3])
    assert contract.get_last_sum() == 6


#
# Test external, and view, methods
#


@pytest.mark.parametrize(
    "method, returndata_expected, return_value_expected",
    [
        ("array", ["0x3", "0x1", "0x2", "0x3"], [1, 2, 3]),
        (
            "array_complex_struct",
            [
                "0x3",
                "0x0",
                "0x7b",
                "0x0",
                "0x0",
                "0x7b",
                "0x1",
                "0x0",
                "0x7b",
                "0x7b",
                "0x0",
                "0x2",
                "0x0",
                "0x0",
                "0x0",
                "0x0",
            ],
            [
                {
                    "timestamp": 0,
                    "value0": 123,
                    "value1": 41854731131275431005995076714107490009088,
                },
                {
                    "timestamp": 1,
                    "value0": 41854731131275431005995076714107490009088,
                    "value1": 123,
                },
                {"timestamp": 2, "value0": 0, "value1": 0},
            ],
        ),
        (
            "array_uint256",
            ["0x3", "0x7b", "0x0", "0x0", "0x7b", "0x0", "0x0"],
            [
                123,
                41854731131275431005995076714107490009088,
                0,
            ],
        ),
        (
            "complex_struct",
            ["0x4d2", "0x7b", "0x0", "0x0", "0x7b"],
            {
                "timestamp": 1234,
                "value0": 123,
                "value1": 41854731131275431005995076714107490009088,
            },
        ),
        ("felt", ["0x2"], 2),
        (
            "mix",
            [
                "0x1",
                "0x2",
                "0x3",
                "0x4",
                "0x7b",
                "0x0",
                "0x3",
                "0x8",
                "0x9",
                "0xa",
                "0xb",
                "0x0",
                "0x7b",
            ],
            (
                1,
                [3, 4],
                123,
                [8, 9, 10],
                11,
                41854731131275431005995076714107490009088,
            ),
        ),
        ("uint256", ["0x1", "0x0"], 1),
        (
            "uint256s",
            ["0x7b", "0x0", "0x0", "0x7b", "0x0", "0x0"],
            (
                123,
                41854731131275431005995076714107490009088,
                0,
            ),
        ),
    ],
)
def test_external_and_view_method_outputs(
    method, returndata_expected, return_value_expected, contract, account
):
    # Check the view method
    return_value = getattr(contract, f"{method}_view")()
    assert return_value == return_value_expected

    # Check the external method
    receipt = getattr(contract, f"{method}_external")(sender=account)
    assert receipt.returndata == returndata_expected
    assert receipt.return_value == return_value



from ape.exceptions import ApeException, ProviderError


class StarknetEcosystemError(ApeException):
    """
    A general Starknet ecosystem error.
    """


class StarknetProviderError(ProviderError):
    """
    An error raised by the Starknet provider.
    """



from typing import Any

from ape.api import ConverterAPI
from ape.types import AddressType
from eth_typing import HexAddress, HexStr

from ape_starknet.utils import PLUGIN_NAME, is_checksum_address, is_hex_address, to_checksum_address


# NOTE: This utility converter ensures that all bytes args can accept hex too
class StarknetAddressConverter(ConverterAPI):
    """
    A converter that converts ``str`` to ``AddressType``.
    """

    def is_convertible(self, value: Any) -> bool:
        provider = self.network_manager.active_provider
        return (
            provider is not None
            and provider.network.ecosystem.name == PLUGIN_NAME
            and isinstance(value, str)
            and is_hex_address(value)
        )

    def convert(self, value: str) -> AddressType:
        """
        Convert the given value to a ``AddressType``.

        Args:
            value (str): The address ``str`` to convert.

        Returns:
            ``AddressType``
        """

        if is_checksum_address(value):
            return AddressType(HexAddress(HexStr(value)))

        return to_checksum_address(value)



from ape.contracts.base import ContractConstructor


def test_deploy_txn_hash(project, convert, provider):
    contract_type = project.MyContract.contract_type
    constructor = ContractConstructor(
        abi=contract_type.constructor, deployment_bytecode=contract_type.get_deployment_bytecode()
    )
    deploy_txn = constructor.serialize_transaction()
    receipt = provider.send_transaction(deploy_txn)

    # Ensure pre-calculated hash equals client response hash.
    assert deploy_txn.txn_hash.hex() == receipt.txn_hash



import pytest

AMOUNT_0 = 10
AMOUNT_1 = 2**128 + 42


@pytest.fixture
def receipt(token_contract, account, second_account):
    return token_contract.fire_events(second_account.address, AMOUNT_0, AMOUNT_1, sender=account)


def test_decode_logs(receipt, token_contract, account, second_account):
    transfer_logs = list(receipt.decode_logs(token_contract.Transfer))
    assert len(transfer_logs) == 1
    log = transfer_logs[0]
    assert log.from_ == int(account.address, 16)
    assert log.to == int(second_account.address, 16)
    assert log.value == AMOUNT_0

    mint_logs = list(receipt.decode_logs(token_contract.Mint))
    assert len(mint_logs) == 1
    log = mint_logs[0]
    assert log.sender == int(account.address, 16)
    assert log.amount0 == AMOUNT_0
    assert log.amount1 == AMOUNT_1
    assert log.to == int(second_account.address, 16)


def test_decode_logs_no_specify_abi(receipt, account, second_account):
    logs = list(receipt.decode_logs())
    assert len(logs) == 2
    transfer_log, mint_log = logs

    assert transfer_log.from_ == int(account.address, 16)
    assert transfer_log.to == int(second_account.address, 16)
    assert transfer_log.value == AMOUNT_0
    assert mint_log.sender == int(account.address, 16)
    assert mint_log.amount0 == AMOUNT_0
    assert mint_log.amount1 == AMOUNT_1
    assert mint_log.to == int(second_account.address, 16)



import pytest

all_tokens = pytest.mark.parametrize(
    "token",
    (
        "eth",
        "test_token",
        # "proxy_token",  # TODO: Fix local proxies
    ),
)


@all_tokens
def test_get_balance(tokens, account, token_initial_supply, token):
    if token == "eth":
        # Likely spent fees
        assert tokens.get_balance(account)
    else:
        assert tokens.get_balance(account, token=token) == token_initial_supply


@all_tokens
def test_transfer(tokens, account, second_account, token):
    initial_balance = tokens.get_balance(second_account.address, token=token)
    tokens.transfer(account.address, second_account.address, 10, token=token)
    assert tokens.get_balance(second_account.address, token=token) == initial_balance + 10


def test_large_transfer(tokens, account, second_account):
    initial_balance = tokens.get_balance(second_account.address, token="test_token")

    # Value large enough to properly test Uint256 logic
    balance_to_transfer = 2**128 + 1
    tokens.transfer(
        account.address, second_account.address, balance_to_transfer, token="test_token"
    )
    actual = tokens.get_balance(second_account.address, token="test_token")
    expected = initial_balance + balance_to_transfer
    assert actual == expected



from typing import List, cast

import click
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.cli import (
    NetworkBoundCommand,
    Path,
    ape_cli_context,
    existing_alias_argument,
    network_option,
)
from ape.cli.options import ApeCliContextObject
from ape.utils import add_padding_to_strings

from ape_starknet.accounts import (
    BaseStarknetAccount,
    StarknetAccountContracts,
    StarknetKeyfileAccount,
)
from ape_starknet.utils import PLUGIN_NAME


def _get_container(cli_ctx: ApeCliContextObject) -> StarknetAccountContracts:
    return cast(StarknetAccountContracts, cli_ctx.account_manager.containers[PLUGIN_NAME])


@click.group("accounts")
def accounts():
    """Manage Starknet accounts"""


@accounts.command(cls=NetworkBoundCommand)
@ape_cli_context()
@click.argument("alias")
@network_option(ecosystem=PLUGIN_NAME)
@click.option("--token", help="Used for deploying contracts in Alpha MainNet.")
def create(cli_ctx, alias, network, token):
    """Deploy an account"""
    container = _get_container(cli_ctx)

    if alias in container.aliases:
        # Check if we have already deployed in this network.
        # The user will have to delete before they can deploy again.
        existing_account = container.load(alias)
        if existing_account.get_deployment(network):
            cli_ctx.abort(
                f"Account already deployed to '{network}' network. "
                f"Run 'ape starknet accounts delete <alias> --network {network}' "
                "first to re-deploy."
            )

    contract_address = container.deploy_account(alias, token=token)
    contract_address = click.style(contract_address, bold=True)
    cli_ctx.logger.success(f"Account successfully deployed to '{contract_address}'.")


@accounts.command("list")
@ape_cli_context()
def _list(cli_ctx):
    """List your Starknet accounts"""

    starknet_accounts = [
        k
        for k in cast(List[StarknetKeyfileAccount], list(_get_container(cli_ctx).accounts))
        if isinstance(k, StarknetKeyfileAccount)
    ]

    if len(starknet_accounts) == 0:
        cli_ctx.logger.warning("No accounts found.")
        return

    num_accounts = len(starknet_accounts)
    header = f"Found {num_accounts} account"
    header += "s" if num_accounts > 1 else ""
    click.echo(f"{header}\n")

    for index in range(num_accounts):
        account = starknet_accounts[index]
        output_dict = {"Alias": account.alias, "Public key": account.address}
        for deployment in account.get_deployments():
            key = f"Contract address ({deployment.network_name})"
            output_dict[key] = deployment.contract_address

        output_keys = add_padding_to_strings(list(output_dict.keys()))
        output_dict = {k: output_dict[k.rstrip()] for k in output_keys}

        for k, v in output_dict.items():
            click.echo(f"{k} - {v}")

        if index < num_accounts - 1:
            click.echo()


@accounts.command(name="import")
@ape_cli_context()
@click.argument("alias")
@network_option(ecosystem=PLUGIN_NAME)
@click.option(
    "--address",
    help="The contract address of the account",
    callback=lambda ctx, param, value: ctx.obj.network_manager.starknet.decode_address(value)
    if value
    else None,
)
@click.option("--keyfile", help="Import an existing key-file", type=Path())
def _import(cli_ctx, alias, network, address, keyfile):
    """Add an existing account"""

    container = _get_container(cli_ctx)

    with cli_ctx.network_manager.parse_network_choice(network) as provider:
        network_name = provider.network.name

    if network_name == LOCAL_NETWORK_NAME:
        cli_ctx.abort("Must use --network option to specify non-local network.")

    if alias in container.aliases:
        existing_account = container.load(alias)

        if existing_account.get_deployment(network_name) or not isinstance(
            existing_account, StarknetKeyfileAccount
        ):
            cli_ctx.abort(f"Account already imported with '{network_name}' network.")

        click.echo(f"Importing existing account to network '{network_name}'.")
        existing_account.add_deployment(network_name, address)

    elif keyfile:
        container.import_account_from_key_file(alias, keyfile)
    elif address:
        private_key = click.prompt("Enter private key", hide_input=True)
        container.import_account(alias, network_name, address, private_key)
    else:
        cli_ctx.abort("Please provide either --keyfile or --address to import this account.")

    cli_ctx.logger.success(f"Import account '{alias}'.")


@accounts.command()
@ape_cli_context()
@existing_alias_argument(account_type=BaseStarknetAccount)
@network_option(ecosystem=PLUGIN_NAME)
def delete(cli_ctx, alias, network):
    """Delete an existing account"""
    container = _get_container(cli_ctx)

    if network == "starknet":
        # Did not specify a network and should use normally use default
        # However, if the account only exists on a single network, assume that one.
        account = container.load(alias)
        deployments = account.get_deployments()
        if len(deployments) == 1:
            network = deployments[0].network_name

    container.delete_account(alias, network=network)
    cli_ctx.logger.success(f"Account '{alias}' on network '{network}' has been deleted.")


@accounts.command()
@ape_cli_context()
@existing_alias_argument(account_type=StarknetKeyfileAccount)
def change_password(cli_ctx, alias):
    """Change the password of an existing account"""
    account = cli_ctx.account_manager.load(alias)
    account.change_password()
    cli_ctx.logger.success(f"Password has been changed for account '{alias}'")



import json
import os
from pathlib import Path
from tempfile import mkdtemp
from typing import cast

import ape
import pytest
from ape.api.networks import LOCAL_NETWORK_NAME, EcosystemAPI
from ape.contracts import ContractContainer
from ethpm_types import ContractType

from ape_starknet import tokens as _tokens
from ape_starknet.accounts import StarknetAccountContracts, StarknetKeyfileAccount
from ape_starknet.utils import PLUGIN_NAME

# NOTE: Ensure that we don't use local paths for these
ape.config.DATA_FOLDER = Path(mkdtemp()).resolve()
ape.config.PROJECT_FOLDER = Path(mkdtemp()).resolve()

_HERE = Path(__file__).parent
projects_directory = Path(__file__).parent / "projects"
project_names = [p.stem for p in projects_directory.iterdir() if p.is_dir()]
ALIAS = "__TEST_ALIAS__"
SECOND_ALIAS = "__TEST_ALIAS_2__"
EXISTING_KEY_FILE_ALIAS = f"{ALIAS}existing_key_file"
EXISTING_EPHEMERAL_ALIAS = f"{ALIAS}existing_ephemeral"
PASSWORD = "123"
PUBLIC_KEY = "0x140dfbab0d711a23dd58842be2ee16318e3de1c7"
CONTRACT_ADDRESS = "0x6b7243AA4edbe5BD629c6712B3aC9639B160480A7730A31483F7B387463a183"

# Purposely pick a number larest enough to test Uint256 logic
TOKEN_INITIAL_SUPPLY = 2 * 2**128

ETH_CONTRACT_TYPE = {
    "abi": [
        {
            "anonymous": False,
            "inputs": [
                {"indexed": False, "name": "prevNum", "type": "uint256"},
                {"indexed": True, "name": "newNum", "type": "uint256"},
            ],
            "name": "NumberChange",
            "type": "event",
        },
        {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
        {
            "inputs": [{"name": "num", "type": "uint256"}],
            "name": "setNumber",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "owner",
            "outputs": [{"name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "myNumber",
            "outputs": [{"name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "prevNumber",
            "outputs": [{"name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
    ],
    "contractName": "EthContract",
    "deploymentBytecode": {
        "bytecode": "0x3360005561012261001c6300000000396101226000016300000000f3600436101561000d57610117565b60003560e01c3461011d57633fb5c1cb81186100d057600054331461008957600b6040527f21617574686f72697a65640000000000000000000000000000000000000000006060526040506040518060600181600003601f1636823750506308c379a06000526020602052601f19601f6040510116604401601cfd5b60056004351461011d576001546002556004356001556004357f2295d5ec33e3af0d43cc4b73aa3cd7d784150fe365cbdb4b4fd338220e4f135760025460405260206040a2005b638da5cb5b81186100e75760005460405260206040f35b6323fd0e4081186100fe5760015460405260206040f35b634825cf6f81186101155760025460405260206040f35b505b60006000fd5b600080fd"  # noqa: E501
    },
    "devdoc": {},
    "runtimeBytecode": {
        "bytecode": "0x600436101561000d57610117565b60003560e01c3461011d57633fb5c1cb81186100d057600054331461008957600b6040527f21617574686f72697a65640000000000000000000000000000000000000000006060526040506040518060600181600003601f1636823750506308c379a06000526020602052601f19601f6040510116604401601cfd5b60056004351461011d576001546002556004356001556004357f2295d5ec33e3af0d43cc4b73aa3cd7d784150fe365cbdb4b4fd338220e4f135760025460405260206040a2005b638da5cb5b81186100e75760005460405260206040f35b6323fd0e4081186100fe5760015460405260206040f35b634825cf6f81186101155760025460405260206040f35b505b60006000fd5b600080fd"  # noqa: E501
    },
    "sourceId": "EthContract.vy",
    "userdoc": {},
}


@pytest.fixture(scope="session")
def config():
    return ape.config


@pytest.fixture(scope="session")
def accounts():
    return ape.accounts


@pytest.fixture(scope="session")
def networks():
    return ape.networks


@pytest.fixture(scope="session")
def convert():
    return ape.convert


@pytest.fixture(scope="session")
def chain():
    return ape.chain


@pytest.fixture(scope="session")
def token_contract(config, account, token_initial_supply, project):
    project_path = projects_directory / "token"

    with config.using_project(project_path):
        return project.get_contract("TestToken").deploy(
            123123, 321321, token_initial_supply, account.address
        )


@pytest.fixture(scope="session")
def proxy_token_contract(config, account, token_initial_supply, token_contract, project):
    project_path = projects_directory / "proxy"

    with config.using_project(project_path):
        contract = project.get_contract("Proxy").deploy(token_contract.address)
        _tokens.add_token("proxy_token", LOCAL_NETWORK_NAME, contract.address)
        return _tokens["proxy_token"]


@pytest.fixture(scope="session")
def tokens(token_contract, proxy_token_contract, provider, account):
    _tokens.add_token("test_token", LOCAL_NETWORK_NAME, token_contract.address)
    return _tokens


@pytest.fixture(scope="session")
def explorer(provider):
    return provider.starknet_explorer


@pytest.fixture(scope="session")
def password():
    return PASSWORD


@pytest.fixture(scope="session")
def public_key():
    return PUBLIC_KEY


@pytest.fixture(scope="session")
def project(request, config):
    project_path = _HERE / "projects" / "project"
    os.chdir(project_path)

    with config.using_project(project_path):
        yield ape.project

    os.chdir(_HERE)


@pytest.fixture(scope="session")
def contract_container(project):
    return project.MyContract


@pytest.fixture(scope="session")
def eth_contract_container(project):
    contract_type = ContractType.parse_obj(ETH_CONTRACT_TYPE)
    return ContractContainer(contract_type)


@pytest.fixture(scope="session")
def contract(account, contract_container, provider):
    deployed_contract = contract_container.deploy()
    deployed_contract.initialize(sender=account)
    return deployed_contract


@pytest.fixture(scope="session")
def factory_contract_container(provider, project):
    return project.ContractFactory


@pytest.fixture
def initial_balance(contract, account):
    return contract.get_balance(account.address)


@pytest.fixture(scope="session")
def provider():
    network = f"{PLUGIN_NAME}:{LOCAL_NETWORK_NAME}:{PLUGIN_NAME}"
    with ape.networks.parse_network_choice(network) as provider:
        yield provider


@pytest.fixture(scope="session")
def account_container(accounts):
    return cast(StarknetAccountContracts, accounts.containers[PLUGIN_NAME])


@pytest.fixture(scope="session")
def account(account_container, provider):
    _ = provider  # Connection required
    return account_container.test_accounts[0]


@pytest.fixture(scope="session")
def second_account(account_container, provider):
    _ = provider  # Connection required
    return account_container.test_accounts[1]


@pytest.fixture(scope="session")
def eth_account(accounts):
    return accounts.test_accounts[0]


@pytest.fixture(scope="session")
def ephemeral_account(account_container, provider):
    _ = provider  # Need connection to deploy account.
    account_container.deploy_account(ALIAS)
    return account_container.load(ALIAS)


@pytest.fixture(scope="session")
def ecosystem(provider) -> EcosystemAPI:
    return provider.network.ecosystem


@pytest.fixture(scope="session")
def key_file_account_data():
    return {
        "address": "140dfbab0d711a23dd58842be2ee16318e3de1c7",
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {"iv": "608494faf88e2d2aea2faac844504233"},
            "ciphertext": "78789f5d4fc5054c18342f020473ecd7c8f75ff2050cdee548121446b40a8ffb",
            "kdf": "scrypt",
            "kdfparams": {
                "dklen": 32,
                "n": 262144,
                "r": 1,
                "p": 8,
                "salt": "c1bbae92537eb53d9ffa3790740bcdb4",
            },
            "mac": "9684c62113753054fc893ffa1b7ae704c65a454f5f89c3b55cc986798d8d5a58",
        },
        "id": "393bb446-55fb-42ec-bd35-ada0b25e17cf",
        "version": 3,
        "ape-starknet": {
            "version": "0.1.0",
            "deployments": [
                {
                    "network_name": "testnet",
                    "contract_address": CONTRACT_ADDRESS,
                },
                {
                    "network_name": "mainnet",
                    "contract_address": CONTRACT_ADDRESS,
                },
            ],
        },
    }


@pytest.fixture(scope="session")
def argent_x_key_file_account_data():
    return {
        "address": "140dfbab0d711a23dd58842be2ee16318e3de1c7",
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {"iv": "608494faf88e2d2aea2faac844504233"},
            "ciphertext": "78789f5d4fc5054c18342f020473ecd7c8f75ff2050cdee548121446b40a8ffb",
            "kdf": "scrypt",
            "kdfparams": {
                "dklen": 32,
                "n": 262144,
                "r": 1,
                "p": 8,
                "salt": "c1bbae92537eb53d9ffa3790740bcdb4",
            },
            "mac": "9684c62113753054fc893ffa1b7ae704c65a454f5f89c3b55cc986798d8d5a58",
        },
        "id": "393bb446-55fb-42ec-bd35-ada0b25e17cf",
        "version": 3,
        "argent": {
            "version": "0.1.0",
            "accounts": [
                {
                    "network": "goerli-alpha",
                    "address": CONTRACT_ADDRESS,
                },
            ],
        },
    }


@pytest.fixture(scope="session")
def ephemeral_account_data():
    return {
        "private_key": 509219664670742235607272813021130138373595301613956902800973975925797957544,
        "public_key": 2068822281043178075870469557539081791152169138879468456959920393634230618024,
    }


@pytest.fixture
def key_file_account(config, key_file_account_data):
    temp_accounts_dir = Path(config.DATA_FOLDER) / "starknet"
    temp_accounts_dir.mkdir(exist_ok=True, parents=True)
    test_key_file_path = temp_accounts_dir / f"{EXISTING_KEY_FILE_ALIAS}.json"

    if test_key_file_path.exists():
        test_key_file_path.unlink()

    test_key_file_path.write_text(json.dumps(key_file_account_data))

    yield StarknetKeyfileAccount(key_file_path=test_key_file_path)

    if test_key_file_path.exists():
        test_key_file_path.unlink()


@pytest.fixture(scope="session")
def token_initial_supply():
    return TOKEN_INITIAL_SUPPLY


@pytest.fixture(scope="session")
def data_folder():
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def traces_testnet_243810(data_folder):
    content = (data_folder / "traces-testnet-block-243810.json").read_text()
    return json.loads(content)["traces"]


@pytest.fixture(scope="session")
def traces_testnet_243810_results(data_folder):
    content = (data_folder / "traces-testnet-block-243810_results.json").read_text()
    return json.loads(content)



import os
from typing import Any, Dict, Iterator, List, Optional, Union
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen

import requests
from ape.api import BlockAPI, ProviderAPI, ReceiptAPI, SubprocessProvider, TransactionAPI
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractInstance
from ape.exceptions import ProviderNotConnectedError, TransactionError
from ape.types import AddressType, BlockID, ContractLog, LogFilter
from ape.utils import DEFAULT_NUMBER_OF_TEST_ACCOUNTS, cached_property, raises_not_implemented
from ethpm_types import ContractType
from starknet_py.net.client_models import (
    BlockSingleTransactionTrace,
    ContractCode,
    InvokeTransaction,
    SentTransactionResponse,
    StarknetBlock,
)
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models import parse_address
from starkware.starkware_utils.error_handling import StarkErrorCode

from ape_starknet.config import DEFAULT_PORT, StarknetConfig
from ape_starknet.exceptions import StarknetEcosystemError, StarknetProviderError
from ape_starknet.tokens import TokenManager
from ape_starknet.transactions import (
    ContractDeclaration,
    InvokeFunctionTransaction,
    StarknetTransaction,
)
from ape_starknet.utils import (
    ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY,
    DEFAULT_ACCOUNT_SEED,
    PLUGIN_NAME,
    extract_trace_data,
    get_chain_id,
    get_dict_from_tx_info,
    get_virtual_machine_error,
    handle_client_errors,
)
from ape_starknet.utils.basemodel import StarknetBase


class StarknetProvider(SubprocessProvider, ProviderAPI, StarknetBase):
    """
    A Starknet provider.
    """

    # Gets set when 'connect()' is called.
    client: Optional[GatewayClient] = None
    token_manager: TokenManager = TokenManager()
    cached_code: Dict[int, ContractCode] = {}

    @property
    def process_name(self) -> str:
        return "starknet-devnet"

    @property
    def is_connected(self) -> bool:
        was_successful = False
        try:
            urlopen(self.uri)
            was_successful = True
        except HTTPError as err:
            was_successful = err.code == 404  # Task failed successfully
        except Exception:
            was_successful = False

        if was_successful and self.client is None:
            self.client = GatewayClient(self.uri, chain=self.chain_id)

        return was_successful

    @property
    def starknet_client(self) -> GatewayClient:
        if not self.is_connected:
            raise StarknetProviderError("Provider is not connected to Starknet.")

        return self.client

    def build_command(self) -> List[str]:
        parts = urlparse(self.uri)
        return [
            self.process_name,
            "--host",
            str(parts.hostname),
            "--port",
            str(parts.port),
            "--accounts",
            str(DEFAULT_NUMBER_OF_TEST_ACCOUNTS),
            "--seed",
            str(DEFAULT_ACCOUNT_SEED),
        ]

    @cached_property
    def plugin_config(self) -> StarknetConfig:
        return self.config_manager.get_config(PLUGIN_NAME) or StarknetConfig()  # type: ignore

    @cached_property
    def uri(self) -> str:
        network_config = self.plugin_config.provider.dict().get(self.network.name)
        if not network_config:
            raise StarknetProviderError(f"Unknown network '{self.network.name}'.")

        return network_config.get("uri") or f"http://127.0.0.1:{DEFAULT_PORT}"

    def connect(self):
        if self.network.name == LOCAL_NETWORK_NAME:
            # Behave like a 'SubprocessProvider'
            if not self.is_connected:
                super().connect()

            self.start()

        self.client = GatewayClient(self.uri, chain=self.chain_id)

    def disconnect(self):
        self.client = None
        super().disconnect()

    def update_settings(self, new_settings: dict):
        pass

    @property
    def chain_id(self) -> int:
        return get_chain_id(self.network.name).value

    @handle_client_errors
    def get_balance(self, address: AddressType) -> int:
        account = self.account_contracts[address]
        return self.token_manager.get_balance(account.address)

    @handle_client_errors
    def get_code(self, address: str) -> List[int]:
        # NOTE: Always return truthy value for code so that Ape core works properly
        return self.get_code_and_abi(address).bytecode or [ord(c) for c in "PROXY"]

    @handle_client_errors
    def get_abi(self, address: str) -> List[Dict]:
        return self.get_code_and_abi(address).abi

    @handle_client_errors
    def get_nonce(self, address: AddressType) -> int:
        # Check if passing a public-key address of a local account
        if address in self.account_contracts.public_key_addresses:
            contract_address = self.account_contracts.get_account(address).address
            if contract_address:
                address = contract_address

        checksum_address = self.starknet.decode_address(address)
        contract = self.chain_manager.contracts.instance_at(checksum_address)

        if not isinstance(contract, ContractInstance):
            raise StarknetProviderError(f"Account contract '{checksum_address}' not found.")

        return contract.get_nonce()

    @handle_client_errors
    def estimate_gas_cost(self, txn: StarknetTransaction) -> int:
        if not self.client:
            raise ProviderNotConnectedError()

        starknet_object = txn.as_starknet_object()
        estimated_fee = self.client.estimate_fee_sync(starknet_object)
        return estimated_fee.overall_fee

    @property
    def gas_price(self) -> int:
        """
        **NOTE**: Currently, the gas price is fixed to always be 100 gwei.
        """
        return self.conversion_manager.convert("100 gwei", int)

    @handle_client_errors
    def get_block(self, block_id: BlockID) -> BlockAPI:
        if isinstance(block_id, (int, str)) and len(str(block_id)) == 76:
            kwarg = "block_hash"
        elif block_id in ("pending", "latest"):
            kwarg = "block_number"
        elif isinstance(block_id, int):
            kwarg = "block_number"
            if block_id < 0:
                latest_block_number = self.get_block("latest").number
                block_id_int = latest_block_number + block_id + 1
                if block_id_int < 0:
                    raise StarknetProviderError(f"Block with number '{block_id_int}' not found.")

                block_id = block_id_int

        else:
            raise StarknetProviderError(f"Unsupported BlockID type '{type(block_id)}'.")

        block = self.starknet_client.get_block_sync(**{kwarg: block_id})
        return self.starknet.decode_block(block)

    def _get_block(self, block_id: BlockID) -> StarknetBlock:
        kwarg = (
            "block_hash"
            if isinstance(block_id, (int, str)) and len(str(block_id)) == 76
            else "block_number"
        )
        return self.starknet_client.get_block_sync(**{kwarg: block_id})

    @handle_client_errors
    def send_call(self, txn: TransactionAPI) -> bytes:
        if not isinstance(txn, InvokeFunctionTransaction):
            type_str = f"{txn.type!r}" if isinstance(txn.type, bytes) else str(txn.type)
            raise StarknetProviderError(
                f"Transaction must be from an invocation. Received type {type_str}."
            )

        if not self.client:
            raise ProviderNotConnectedError()

        starknet_obj = txn.as_starknet_object()
        return self.client.call_contract_sync(starknet_obj)  # type: ignore

    @handle_client_errors
    def _get_traces(self, block_number: int) -> List[BlockSingleTransactionTrace]:
        block_traces = self.starknet_client.get_block_traces_sync(block_number=block_number)
        return block_traces.traces

    @handle_client_errors
    def _get_single_trace(
        self, block_number: int, txn_hash: int
    ) -> Optional[BlockSingleTransactionTrace]:
        traces = self._get_traces(block_number)
        return next((trace for trace in traces if trace.transaction_hash == txn_hash), None)

    @handle_client_errors
    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        self.starknet_client.wait_for_tx_sync(txn_hash)
        txn_info = self.starknet_client.get_transaction_sync(tx_hash=txn_hash)
        receipt = self.starknet_client.get_transaction_receipt_sync(tx_hash=txn_hash)

        trace_data = {}
        if isinstance(txn_info, InvokeTransaction):
            trace = self._get_single_trace(receipt.block_number, receipt.hash)
            trace_data = extract_trace_data(trace)

        receipt_dict: Dict[str, Any] = {"provider": self, **trace_data, **vars(receipt)}
        receipt_dict = get_dict_from_tx_info(txn_info, **receipt_dict)
        return self.starknet.decode_receipt(receipt_dict)

    def get_transactions_by_block(self, block_id: BlockID) -> Iterator[TransactionAPI]:
        block = self._get_block(block_id)
        for txn_info in block.transactions:
            txn_dict = get_dict_from_tx_info(txn_info)
            yield self.starknet.create_transaction(**txn_dict)

    @handle_client_errors
    def send_transaction(self, txn: TransactionAPI, token: Optional[str] = None) -> ReceiptAPI:
        response = self._send_transaction(txn, token=token)
        if response.code != StarkErrorCode.TRANSACTION_RECEIVED.name:
            raise TransactionError(message="Transaction not received.")

        receipt = self.get_transaction(response.transaction_hash)

        if isinstance(txn, InvokeFunctionTransaction):
            if txn.original_method_abi:
                # Use ABI before going through account contract
                abi = txn.original_method_abi
            else:
                abi = txn.method_abi

            return_value = self.starknet.decode_returndata(abi, receipt.returndata)
            receipt.return_value = return_value

        return receipt

    def _send_transaction(
        self, txn: TransactionAPI, token: Optional[str] = None
    ) -> SentTransactionResponse:
        txn = self.prepare_transaction(txn)
        if not token and hasattr(txn, "token") and txn.token:  # type: ignore
            token = txn.token  # type: ignore
        else:
            token = os.environ.get(ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY)

        if not isinstance(txn, StarknetTransaction):
            raise StarknetEcosystemError(
                "Unable to send non-Starknet transaction using a Starknet provider."
            )

        starknet_txn = txn.as_starknet_object()
        return self.starknet_client.send_transaction_sync(starknet_txn, token=token)

    @raises_not_implemented
    def get_contract_logs(self, log_filter: LogFilter) -> Iterator[ContractLog]:
        pass

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        return txn

    def set_timestamp(self, new_timestamp: int):
        pending_timestamp = self.get_block("pending").timestamp
        seconds_to_increase = new_timestamp - pending_timestamp
        response = requests.post(
            url=f"{self.uri}/increase_time", json={"time": seconds_to_increase}
        )
        response.raise_for_status()
        response_data = response.json()
        if "timestamp_increased_by" not in response_data:
            raise StarknetProviderError(response_data)

    def get_virtual_machine_error(self, exception: Exception):
        return get_virtual_machine_error(exception)

    def get_code_and_abi(self, address: Union[str, AddressType, int]) -> ContractCode:
        address_int = parse_address(address)

        # Cache code for faster look-up
        if address_int not in self.cached_code:
            self.cached_code[address_int] = self.starknet_client.get_code_sync(address_int)

        return self.cached_code[address_int]

    @handle_client_errors
    def declare(self, contract_type: ContractType) -> ContractDeclaration:
        transaction = self.starknet.encode_contract_declaration(contract_type)
        return self.provider.send_transaction(transaction)


__all__ = ["StarknetProvider"]



from enum import Enum
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, Union

from ape.api import BlockAPI, EcosystemAPI, ReceiptAPI, TransactionAPI
from ape.api.networks import ProxyInfoAPI
from ape.contracts import ContractContainer
from ape.types import AddressType, ContractLog, RawAddress
from eth_utils import is_0x_prefixed
from ethpm_types import ContractType
from ethpm_types.abi import ConstructorABI, EventABI, EventABIType, MethodABI
from hexbytes import HexBytes
from starknet_py.constants import OZ_PROXY_STORAGE_KEY
from starknet_py.net.client_models import StarknetBlock
from starknet_py.net.models.address import parse_address
from starknet_py.net.models.chains import StarknetChainId
from starknet_py.utils.data_transformer.data_transformer import CairoSerializer
from starkware.starknet.definitions.fields import ContractAddressSalt
from starkware.starknet.definitions.transaction_type import TransactionType
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.public.abi_structs import identifier_manager_from_abi
from starkware.starknet.services.api.contract_class import ContractClass

from ape_starknet.exceptions import StarknetEcosystemError, StarknetProviderError
from ape_starknet.transactions import (
    ContractDeclaration,
    DeclareTransaction,
    DeployReceipt,
    DeployTransaction,
    InvocationReceipt,
    InvokeFunctionTransaction,
    StarknetReceipt,
    StarknetTransaction,
)
from ape_starknet.utils import to_checksum_address
from ape_starknet.utils.basemodel import StarknetBase

NETWORKS = {
    # chain_id, network_id
    "mainnet": (StarknetChainId.MAINNET.value, StarknetChainId.MAINNET.value),
    "testnet": (StarknetChainId.TESTNET.value, StarknetChainId.TESTNET.value),
}


class ProxyType(Enum):
    LEGACY = 0
    ARGENT_X = 1
    OPEN_ZEPPELIN = 2


class StarknetProxy(ProxyInfoAPI):
    """
    A proxy contract in Starknet.
    """

    type: ProxyType


class Starknet(EcosystemAPI, StarknetBase):
    """
    The Starknet ``EcosystemAPI`` implementation.
    """

    proxy_info_cache: Dict[AddressType, Optional[StarknetProxy]] = {}

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @classmethod
    def decode_address(cls, raw_address: RawAddress) -> AddressType:
        """
        Make a checksum address given a supported format.
        Borrowed from ``eth_utils.to_checksum_address()`` but supports
        non-length 42 addresses.
        Args:
            raw_address (Union[int, str, bytes]): The value to convert.
        Returns:
            ``AddressType``: The converted address.
        """
        return to_checksum_address(raw_address)

    @classmethod
    def encode_address(cls, address: Union[AddressType, str]) -> int:
        return parse_address(address)

    def serialize_transaction(self, transaction: TransactionAPI) -> bytes:
        if not isinstance(transaction, StarknetTransaction):
            raise StarknetEcosystemError(f"Can only serialize '{StarknetTransaction.__name__}'.")

        starknet_object = transaction.as_starknet_object()
        return starknet_object.deserialize()

    def decode_returndata(self, abi: MethodABI, raw_data: List[int]) -> Any:  # type: ignore
        if not raw_data:
            return raw_data

        full_abi = [
            a.dict() for a in (abi.contract_type.abi if abi.contract_type is not None else [abi])
        ]
        transformer = CairoSerializer(identifier_manager_from_abi(full_abi))

        raw_data = [self.encode_primitive_value(v) for v in raw_data]
        decoded = transformer.to_python(abi.dict()["outputs"], raw_data)

        # Keep only the expected data instead of a 1-item array
        if len(abi.outputs) == 1 or (
            len(abi.outputs) == 2 and str(abi.outputs[1].type).endswith("*")
        ):
            decoded = decoded[0]

        return decoded

    def encode_calldata(
        self,
        full_abi: List,
        method_abi: Union[ConstructorABI, MethodABI],
        call_args: Union[List, Tuple],
    ) -> List:
        full_abi = [abi.dict() if hasattr(abi, "dict") else abi for abi in full_abi]
        transformer = CairoSerializer(identifier_manager_from_abi(full_abi))
        pre_encoded_args: List[Any] = []
        index = 0
        last_index = min(len(method_abi.inputs), len(call_args)) - 1
        did_process_array_during_arr_len = False

        for call_arg, input_type in zip(call_args, method_abi.inputs):
            if str(input_type.type).endswith("*"):
                if did_process_array_during_arr_len:
                    did_process_array_during_arr_len = False
                    continue

                encoded_arg = self._pre_encode_value(call_arg)
                pre_encoded_args.append(encoded_arg)
            elif (
                input_type.name is not None
                and input_type.name.endswith("_len")
                and index < last_index
                and str(method_abi.inputs[index + 1].type).endswith("*")
            ):
                pre_encoded_arg = self._pre_encode_value(call_arg)

                if isinstance(pre_encoded_arg, int):
                    # 'arr_len' was provided.
                    array_index = index + 1
                    pre_encoded_array = self._pre_encode_array(call_args[array_index])
                    pre_encoded_args.append(pre_encoded_array)
                    did_process_array_during_arr_len = True
                else:
                    pre_encoded_args.append(pre_encoded_arg)

            else:
                pre_encoded_args.append(self._pre_encode_value(call_arg))

            index += 1

        calldata, _ = transformer.from_python(method_abi.dict()["inputs"], *pre_encoded_args)
        return calldata

    def _pre_encode_value(self, value: Any) -> Any:
        if isinstance(value, dict):
            return self._pre_encode_struct(value)
        elif isinstance(value, (list, tuple)):
            return self._pre_encode_array(value)
        else:
            return self.encode_primitive_value(value)

    def _pre_encode_array(self, array: Any) -> List:
        if not isinstance(array, (list, tuple)):
            # Will handle single item structs and felts.
            return self._pre_encode_array([array])

        encoded_array = []
        for item in array:
            encoded_value = self._pre_encode_value(item)
            encoded_array.append(encoded_value)

        return encoded_array

    def _pre_encode_struct(self, struct: Dict) -> Dict:
        encoded_struct = {}
        for key, value in struct.items():
            encoded_struct[key] = self._pre_encode_value(value)

        return encoded_struct

    def encode_primitive_value(self, value: Any) -> int:
        if isinstance(value, int):
            return value

        elif isinstance(value, str) and is_0x_prefixed(value):
            return int(value, 16)

        elif isinstance(value, HexBytes):
            return int(value.hex(), 16)

        return value

    def decode_receipt(self, data: dict) -> ReceiptAPI:
        txn_type = TransactionType(data["type"])
        receipt_cls: Type[StarknetReceipt]
        if txn_type == TransactionType.INVOKE_FUNCTION:
            receipt_cls = InvocationReceipt
        elif txn_type == TransactionType.DEPLOY:
            receipt_cls = DeployReceipt
        elif txn_type == TransactionType.DECLARE:
            receipt_cls = ContractDeclaration
        else:
            raise StarknetProviderError(f"Unable to handle contract type '{txn_type.value}'.")

        receipt = receipt_cls.parse_obj(data)

        if receipt is None:
            raise StarknetProviderError("Failed to parse receipt from data.")

        return receipt

    def decode_block(self, block: StarknetBlock) -> BlockAPI:
        return BlockAPI(
            hash=HexBytes(block.block_hash),
            number=block.block_number,
            parentHash=HexBytes(block.parent_block_hash),
            size=len(block.transactions),  # TODO: Figure out size
            timestamp=block.timestamp,
        )

    def encode_deployment(
        self, deployment_bytecode: HexBytes, abi: ConstructorABI, *args, **kwargs
    ) -> TransactionAPI:
        salt = kwargs.get("salt")
        if not salt:
            salt = ContractAddressSalt.get_random_value()

        constructor_args = list(args)
        contract = ContractClass.deserialize(deployment_bytecode)
        calldata = self.encode_calldata(contract.abi, abi, constructor_args)
        return DeployTransaction(
            salt=salt,
            constructor_calldata=calldata,
            contract_code=contract.dumps(),
            token=kwargs.get("token"),
        )

    def encode_transaction(
        self, address: AddressType, abi: MethodABI, *args, **kwargs
    ) -> TransactionAPI:
        # NOTE: This method only works for invoke-transactions
        contract_type = self.starknet_explorer.get_contract_type(address)
        if not contract_type:
            raise StarknetProviderError(f"No contract found at address '{address}'.")

        encoded_calldata = self.encode_calldata(contract_type.abi, abi, list(args))

        return InvokeFunctionTransaction(
            contract_address=address,
            method_abi=abi,
            calldata=encoded_calldata,
            sender=kwargs.get("sender"),
            max_fee=kwargs.get("max_fee"),
        )

    def encode_contract_declaration(
        self, contract: Union[ContractContainer, ContractType], *args, **kwargs
    ) -> DeclareTransaction:
        contract_type = (
            contract.contract_type if isinstance(contract, ContractContainer) else contract
        )
        code = (
            (contract_type.deployment_bytecode.bytecode or 0)
            if contract_type.deployment_bytecode
            else 0
        )
        starknet_contract = ContractClass.deserialize(HexBytes(code))
        return DeclareTransaction(contract_type=contract_type, data=starknet_contract.dumps())

    def create_transaction(self, **kwargs) -> TransactionAPI:
        txn_type = TransactionType(kwargs.pop("type", kwargs.pop("tx_type", "")))
        txn_cls: Union[
            Type[InvokeFunctionTransaction], Type[DeployTransaction], Type[DeclareTransaction]
        ]
        invoking = txn_type == TransactionType.INVOKE_FUNCTION
        if invoking:
            txn_cls = InvokeFunctionTransaction
        elif txn_type == TransactionType.DEPLOY:
            txn_cls = DeployTransaction
        elif txn_type == TransactionType.DECLARE:
            txn_cls = DeclareTransaction

        txn_data: Dict[str, Any] = {**kwargs, "signature": None}
        if "chain_id" not in txn_data and self.network_manager.active_provider:
            txn_data["chain_id"] = self.provider.chain_id

        # For deploy-txns, 'contract_address' is the address of the newly deployed contract.
        if "contract_address" in txn_data:
            txn_data["contract_address"] = self.decode_address(txn_data["contract_address"])

        if not invoking:
            return txn_cls(**txn_data)

        """ ~ Invoke transactions ~ """

        if "method_abi" not in txn_data:
            contract_int = txn_data["contract_address"]
            contract_str = self.decode_address(contract_int)
            contract = self.chain_manager.contracts.get(contract_str)
            if not contract:
                raise StarknetEcosystemError(
                    "Unable to create transaction objects from other networks."
                )

            selector = txn_data["entry_point_selector"]
            if isinstance(selector, str):
                selector = int(selector, 16)

            for abi in contract.mutable_methods:
                selector_to_check = get_selector_from_name(abi.name)

                if selector == selector_to_check:
                    txn_data["method_abi"] = abi

        if "calldata" in txn_data and txn_data["calldata"] is not None:
            # Transactions in blocks show calldata as flattened hex-strs
            # but elsewhere we expect flattened ints. Convert to ints for
            # consistency and testing purposes.
            encoded_calldata = [self.encode_primitive_value(v) for v in txn_data["calldata"]]
            txn_data["calldata"] = encoded_calldata

        return txn_cls(**txn_data)

    def decode_logs(self, logs: List[Dict], *events: EventABI) -> Iterator["ContractLog"]:
        events_by_selector = {get_selector_from_name(e.name): e for e in events}
        log_map = {s: [log for log in logs if s in log["keys"]] for s in events_by_selector}

        def from_uint(low: int, high: int) -> int:
            return low + (high << 128)

        def decode_items(
            abi_types: List[EventABIType], data: List[int]
        ) -> List[Union[int, Tuple[int, int]]]:
            decoded: List[Union[int, Tuple[int, int]]] = []
            iter_data = iter(data)
            for abi_type in abi_types:
                if abi_type.type == "Uint256":
                    # Uint256 are stored using 2 slots
                    decoded.append(from_uint(next(iter_data), next(iter_data)))
                else:
                    decoded.append(next(iter_data))
            return decoded

        for index, (selector, logs) in enumerate(log_map.items()):
            abi = events_by_selector[selector]
            for log in logs:
                event_args = dict(
                    zip([a.name for a in abi.inputs], decode_items(abi.inputs, log["data"]))
                )
                yield ContractLog(  # type: ignore
                    block_hash=log["block_hash"],
                    block_number=log["block_number"],
                    contract_address=self.decode_address(log["from_address"]),
                    event_arguments=event_args,
                    event_name=abi.name,
                    log_index=index,
                    transaction_hash=log["transaction_hash"],
                    transaction_index=0,  # Not available
                )

    def get_proxy_info(self, address: AddressType) -> Optional[StarknetProxy]:
        # Proxies are handled elsewhere in Starknet due to ecosystem differences
        # (namely, contract classes function different than proxy contracts in Ethereum).
        return None

    def _get_proxy_info(
        self, address: AddressType, contract_type: ContractType
    ) -> Optional[StarknetProxy]:
        proxy_type: Optional[ProxyType] = None
        target: Optional[int] = None
        instance = self.chain_manager.contracts.instance_at(address, contract_type=contract_type)
        # Legacy proxy check
        if "implementation" in contract_type.view_methods:
            target = instance.implementation()  # type: ignore
            proxy_type = ProxyType.LEGACY

        # Argent-X proxy check
        elif "get_implementation" in contract_type.view_methods:
            target = instance.get_implementation()  # type: ignore
            proxy_type = ProxyType.ARGENT_X

        # OpenZeppelin proxy check
        elif self.provider.client is not None:
            address_int = self.encode_address(address)
            target = self.provider.client.get_storage_at_sync(
                contract_address=address_int, key=OZ_PROXY_STORAGE_KEY
            )
            if target == "0x0":
                target = None
            else:
                proxy_type = ProxyType.OPEN_ZEPPELIN

        return (
            StarknetProxy(target=self.decode_address(target), type=proxy_type)
            if target and proxy_type
            else None
        )



import re
from dataclasses import asdict
from typing import Any, Dict, List, Optional, Union

from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractEvent
from ape.exceptions import ApeException, ContractLogicError, OutOfGasError
from ape.types import AddressType, RawAddress
from eth_typing import HexAddress, HexStr
from eth_utils import add_0x_prefix, is_text, remove_0x_prefix
from ethpm_types import ContractType
from ethpm_types.abi import EventABI
from hexbytes import HexBytes
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    BlockSingleTransactionTrace,
    DeclareTransaction,
    DeployTransaction,
    InvokeTransaction,
    Transaction,
)
from starknet_py.net.models import TransactionType
from starknet_py.net.models.address import parse_address
from starknet_py.transaction_exceptions import TransactionRejectedError
from starkware.crypto.signature.fast_pedersen_hash import pedersen_hash
from starkware.crypto.signature.signature import get_random_private_key as get_random_pkey
from starkware.starknet.definitions.general_config import StarknetChainId
from starkware.starknet.services.api.contract_class import ContractClass

from ape_starknet.exceptions import StarknetProviderError

PLUGIN_NAME = "starknet"
NETWORKS = {
    # chain_id, network_id
    "mainnet": (StarknetChainId.MAINNET.value, StarknetChainId.MAINNET.value),
    "testnet": (StarknetChainId.TESTNET.value, StarknetChainId.TESTNET.value),
}
_HEX_ADDRESS_REG_EXP = re.compile("(0x)?[0-9a-f]*", re.IGNORECASE | re.ASCII)
"""Same as from eth-utils except not limited length."""
ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY = "ALPHA_MAINNET_WL_DEPLOY_TOKEN"
DEFAULT_ACCOUNT_SEED = 2147483647  # Prime
ContractEventABI = Union[List[Union[EventABI, ContractEvent]], Union[EventABI, ContractEvent]]


def get_chain_id(network_id: Union[str, int]) -> StarknetChainId:
    if isinstance(network_id, int):
        return StarknetChainId(network_id)

    elif network_id == LOCAL_NETWORK_NAME:
        return StarknetChainId.TESTNET  # Use TESTNET chain ID for local network

    elif network_id not in NETWORKS:
        raise StarknetProviderError(f"Unknown network '{network_id}'.")

    return StarknetChainId(NETWORKS[network_id][0])


def to_checksum_address(address: RawAddress) -> AddressType:
    if isinstance(address, bytes):
        address = HexBytes(address).hex()

    address_int = parse_address(address)
    address_str = pad_hex_str(HexBytes(address_int).hex().lower())
    chars = [c for c in remove_0x_prefix(HexStr(address_str))]
    hashed = [b for b in HexBytes(pedersen_hash(0, address_int))]

    for i in range(0, len(chars), 2):
        try:
            if hashed[i >> 1] >> 4 >= 8:
                chars[i] = chars[i].upper()
            if (hashed[i >> 1] & 0x0F) >= 8:
                chars[i + 1] = chars[i + 1].upper()
        except IndexError:
            continue

    return AddressType(HexAddress(add_0x_prefix(HexStr("".join(chars)))))


def is_hex_address(value: Any) -> bool:
    return _HEX_ADDRESS_REG_EXP.fullmatch(value) is not None if is_text(value) else False


def is_checksum_address(value: Any) -> bool:
    if not is_text(value):
        return False

    if not is_hex_address(value):
        return False

    return value == to_checksum_address(value)


def extract_trace_data(trace: BlockSingleTransactionTrace) -> Dict[str, Any]:
    if not trace:
        return {}

    trace_data = trace.function_invocation

    # Keep the most relevant `result`: given the account implementation, `result`
    # may contain an additional number prepend to the data to expose the total
    # number of items. For a method returning a 3-items array like `[1, 2, 3]`,
    # in such scenario `results` would be `[0x4, 0x3, 0x1, 0x2, 0x3]` (the prepend
    # number: 4, the array length: 3, then array items: 1, 2, and 3).
    # As there is no known way to guess when to remove such a number, we prefer to "scan"
    # trace internals to select the most appropriate result. For instance, when `result`
    # contains the additional value, we just need to use the "internal call" `result`
    # that will contain the exact value the method returned.
    invokation_result = trace_data["result"]
    internal_calls = trace_data["internal_calls"]
    trace_data["result"] = (
        internal_calls[-1]["result"] if internal_calls else invokation_result
    ) or invokation_result

    return trace_data


def handle_client_errors(f):
    def func(*args, **kwargs):
        try:
            result = f(*args, **kwargs)
            if isinstance(result, dict) and result.get("error"):
                message = result["error"].get("message") or "Transaction failed"
                raise StarknetProviderError(message)

            return result

        except Exception as err:
            raise get_virtual_machine_error(err) from err

    return func


def get_virtual_machine_error(err: Exception) -> Optional[Exception]:
    if isinstance(err, ApeException) or not isinstance(
        err, (ClientError, TransactionRejectedError)
    ):
        return err

    err_msg = err.message

    if "Actual fee exceeded max fee" in err_msg:
        return OutOfGasError()

    if isinstance(err, ClientError):
        # Remove https://github.com/software-mansion/starknet.py/blob/0.4.3-alpha/starknet_py/net/client_errors.py#L11 # noqa
        err_msg = err_msg.split(":", 1)[-1]

    if "Error message:" in err_msg:
        err_msg = err_msg.split("Error message:")[-1].splitlines()[0]
        return ContractLogicError(revert_message=err_msg.strip())

    if "Error at pc=" in err_msg:
        return ContractLogicError(revert_message=err_msg.strip())

    return StarknetProviderError(err_msg.strip())


def get_dict_from_tx_info(txn_info: Transaction, **extra_kwargs) -> Dict:
    txn_dict = {**asdict(txn_info), **extra_kwargs}

    if isinstance(txn_info, DeployTransaction):
        txn_dict["contract_address"] = to_checksum_address(txn_info.contract_address)
        txn_dict["max_fee"] = 0
        txn_dict["type"] = TransactionType.DEPLOY
    elif isinstance(txn_info, InvokeTransaction):
        txn_dict["contract_address"] = to_checksum_address(txn_info.contract_address)
        txn_dict["events"] = [vars(e) for e in txn_dict.get("events", [])]
        txn_dict["type"] = TransactionType.INVOKE_FUNCTION
    elif isinstance(txn_info, DeclareTransaction):
        txn_dict["sender"] = to_checksum_address(txn_info.sender_address)
        txn_dict["type"] = TransactionType.DECLARE

    return txn_dict


def convert_contract_class_to_contract_type(contract_class: ContractClass):
    return ContractType.parse_obj(
        {
            "contractName": "Account",
            "sourceId": "openzeppelin.account.Account.cairo",
            "deploymentBytecode": {"bytecode": contract_class.serialize().hex()},
            "runtimeBytecode": {},
            "abi": contract_class.abi,
        }
    )


def get_random_private_key() -> str:
    private_key = HexBytes(get_random_pkey()).hex()
    return pad_hex_str(private_key)


def pad_hex_str(value: str, to_length: int = 66) -> str:
    val = value.replace("0x", "")
    actual_len = len(val)
    padding = "0" * (to_length - 2 - actual_len)
    return f"0x{padding}{val}"



import pytest
from ape.exceptions import ApeException, ContractLogicError, OutOfGasError
from hexbytes import HexBytes
from starknet_py.net.client_errors import ClientError, ContractNotFoundError
from starknet_py.net.client_models import BlockSingleTransactionTrace
from starknet_py.transaction_exceptions import TransactionRejectedError

from ape_starknet.exceptions import StarknetProviderError
from ape_starknet.utils import (
    extract_trace_data,
    get_random_private_key,
    get_virtual_machine_error,
    is_checksum_address,
    to_checksum_address,
)


@pytest.mark.parametrize("iteration", range(10))
def test_get_random_private_key(iteration):
    pkey = get_random_private_key()
    assert len(pkey) == 66
    pkey_int = int(pkey, 16)
    pkey_back_to_str = HexBytes(pkey_int).hex()
    assert pkey_back_to_str.replace("0x", "") in pkey


def test_is_checksum_address(account):
    assert is_checksum_address(account.address)


def test_to_checksum_address(account):
    address_lowered = account.address.lower()
    actual = to_checksum_address(address_lowered)
    assert is_checksum_address(actual)
    assert actual == account.address


@pytest.mark.parametrize(
    "exception, expected",
    [
        (ApeException("Foo!"), ApeException("Foo!")),
        (
            ClientError(
                code=500,
                message=(
                    '{"message":"Error at pc=0:91:\nGot an exception '
                    "while executing a hint.\nCairo traceback (most r"
                    "ecent call last):\nUnknown location (pc=0:739)\n"
                    "Unknown location (pc=0:682)\nUnknown location (p"
                    "c=0:358)\nUnknown location (pc=0:400)\nUnknown l"
                    "ocation (pc=0:423)\n\nError in the called contra"
                    "ct (0x123):\nError at pc=0:41:\nGot an exception"
                    " while executing a hint.\nCairo traceback (most "
                    "recent call last):\nUnknown location (pc=0:7453)"
                    "\nUnknown location (pc=0:7437)\nUnknown location"
                    " (pc=0:4491)\nError message: Strings: exceeding "
                    "max felt string length (31)\nUnknown location (p"
                    "c=0:3219)\nUnknown location (pc=0:47)\n\nTraceba"
                    'ck (most recent call last):\n  File "<hint6>", l'
                    "ine 3, in <module>\nAssertionError: a = 36185027"
                    "886661312136973227830950701056231072153315966999"
                    '20561358720 is out of range.","status_code":500}'
                ),
            ),
            ContractLogicError(revert_message="Strings: exceeding max felt string length (31)"),
        ),
        (
            ContractNotFoundError(block_hash="pending"),
            StarknetProviderError("No contract found for identifier: pending"),
        ),
        (
            TransactionRejectedError(message="Actual fee exceeded max fee.\n999800000000000 > 1"),
            OutOfGasError(),
        ),
        (
            TransactionRejectedError(
                message="Error at pc=0:330:\nAn ASSERT_EQ instruction failed: 0 != 1."
            ),
            ContractLogicError(
                revert_message="Error at pc=0:330:\nAn ASSERT_EQ instruction failed: 0 != 1."
            ),
        ),
        (ValueError("Foo!"), ValueError("Foo!")),
    ],
)
def test_get_virtual_machine_error(exception, expected):
    error = get_virtual_machine_error(exception)
    assert str(error) == str(expected)


def test_extract_trace_data(traces_testnet_243810, traces_testnet_243810_results):
    for trace in traces_testnet_243810:
        trace_object = BlockSingleTransactionTrace(**trace)
        trace_data = extract_trace_data(trace_object)
        assert isinstance(trace_data, dict)

        expected_result = traces_testnet_243810_results[trace_object.transaction_hash]
        assert trace_data["result"] == expected_result



from typing import Iterator, Optional

from ape.api import ExplorerAPI, ReceiptAPI
from ape.types import AddressType
from ethpm_types import ContractType

from ape_starknet.utils.basemodel import StarknetBase


class StarknetExplorer(ExplorerAPI, StarknetBase):
    BASE_URIS = {
        "testnet": "https://goerli.voyager.online",
        "mainnet": "https://voyager.online",
    }

    @property
    def base_uri(self) -> str:
        network_name = self.provider.network.name
        return self.BASE_URIS.get(network_name, "")

    def get_address_url(self, address: AddressType) -> str:
        base_uri = self.base_uri
        return f"{base_uri}/contracts/{address}" if base_uri else ""

    def get_transaction_url(self, transaction_hash: str) -> str:
        base_uri = self.base_uri
        return f"{base_uri}/txns/{transaction_hash}" if base_uri else ""

    def get_contract_type(self, address: AddressType) -> Optional[ContractType]:
        if self.tokens.is_token(address):
            return self.tokens.contract_type

        elif address in self.account_contracts:
            starknet_account = self.account_contracts[address]
            return starknet_account.get_contract_type()  # type: ignore

        code_and_abi = self.provider.get_code_and_abi(address)
        contract_type_dict = {"abi": code_and_abi.abi}
        return ContractType.parse_obj(contract_type_dict)

    def get_account_transactions(self, address: AddressType) -> Iterator[ReceiptAPI]:
        # TODO
        yield from ()



from dataclasses import asdict
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from ape.api import ReceiptAPI, TransactionAPI
from ape.contracts import ContractContainer, ContractInstance
from ape.exceptions import TransactionError
from ape.types import AddressType, ContractLog
from ape.utils import abstractmethod, cached_property
from eth_utils import to_int
from ethpm_types import ContractType, HexBytes
from ethpm_types.abi import EventABI, MethodABI
from pydantic import Field, validator
from starknet_py.net.client_models import Event, TransactionStatus
from starknet_py.net.models.transaction import (
    Declare,
    Deploy,
    InvokeFunction,
    Transaction,
    TransactionType,
)
from starkware.starknet.core.os.class_hash import compute_class_hash
from starkware.starknet.core.os.contract_address.contract_address import calculate_contract_address
from starkware.starknet.core.os.transaction_hash.transaction_hash import (
    TransactionHashPrefix,
    calculate_declare_transaction_hash,
    calculate_deploy_transaction_hash,
    calculate_transaction_hash_common,
)
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.services.api.contract_class import ContractClass
from starkware.starknet.services.api.gateway.transaction import DECLARE_SENDER_ADDRESS
from starkware.starknet.testing.contract_utils import get_contract_class

from ape_starknet.utils import ContractEventABI, to_checksum_address
from ape_starknet.utils.basemodel import StarknetBase


class StarknetTransaction(TransactionAPI, StarknetBase):
    """
    A base transaction class for all Starknet transactions.
    """

    status: int = TransactionStatus.NOT_RECEIVED
    version: int = 0

    """Ignored"""
    gas_limit: int = Field(0, exclude=True)
    max_fee: Optional[int] = Field(None, exclude=True)
    max_priority_fee: Optional[int] = Field(None, exclude=True)

    class Config:
        use_enum_values = True

    def serialize_transaction(self) -> dict:  # type: ignore
        return self.dict()

    @validator("status", pre=True, allow_reuse=True)
    def validate_status(cls, value):
        if isinstance(value, TransactionStatus):
            return value.value

        elif isinstance(value, str):
            return int(value, 16)

        return value

    @abstractmethod
    def as_starknet_object(self) -> Transaction:
        """
        Convert :class:`~ape.api.providers.TransactionAPI` to its Starknet
        transaction equivalent so it can be accepted by the core Starknet OS
        framework.
        """

    @property
    def total_transfer_value(self) -> int:
        max_fee = self.max_fee or 0
        return self.value + max_fee


class DeclareTransaction(StarknetTransaction):
    sender: AddressType = to_checksum_address(DECLARE_SENDER_ADDRESS)
    type: TransactionType = TransactionType.DECLARE

    @property
    def starknet_contract(self) -> ContractClass:
        return ContractClass.deserialize(self.data)

    @property
    def txn_hash(self) -> HexBytes:
        return calculate_declare_transaction_hash(
            self.starknet_contract,
            self.provider.chain_id,
            self.sender,
        )

    def as_starknet_object(self) -> Declare:
        sender_int = self.starknet.encode_address(self.sender)

        # NOTE: The sender address is a special address, nonce has to be 0, and signatures
        # and fees are not supported.
        return Declare(
            contract_class=self.starknet_contract,
            max_fee=0,
            nonce=0,
            sender_address=sender_int,
            signature=[],
            version=self.version,
        )


class DeployTransaction(StarknetTransaction):
    salt: int

    caller_address: int = 0
    constructor_calldata: Union[List, Tuple] = []
    token: Optional[str] = None
    type: TransactionType = TransactionType.DEPLOY

    """Aliases"""
    data: bytes = Field(alias="contract_code")

    """Ignored"""
    receiver: Optional[str] = Field(None, exclude=True)

    @property
    def starknet_contract(self) -> Optional[ContractClass]:
        return ContractClass.deserialize(self.data)

    @property
    def txn_hash(self) -> HexBytes:
        contract_address = calculate_contract_address(
            contract_class=self.starknet_contract,
            constructor_calldata=self.constructor_calldata,
            deployer_address=self.caller_address,
            salt=self.salt,
        )
        hash_int = calculate_deploy_transaction_hash(
            chain_id=self.provider.chain_id,
            contract_address=contract_address,
            constructor_calldata=self.constructor_calldata,
            version=self.version,
        )
        return HexBytes(hash_int)

    def as_starknet_object(self) -> Deploy:
        return Deploy(
            constructor_calldata=self.constructor_calldata,
            contract_address_salt=self.salt,
            contract_definition=self.starknet_contract,
            version=self.version,
        )


class InvokeFunctionTransaction(StarknetTransaction):
    max_fee: Optional[int] = None
    method_abi: MethodABI

    sender: Optional[AddressType] = None
    type: TransactionType = TransactionType.INVOKE_FUNCTION

    original_method_abi: Optional[MethodABI] = None
    """
    Only set when invoked from an account `__execute__`
    special method to help decoding return data
    """

    """Aliases"""
    data: List[Any] = Field(alias="calldata")  # type: ignore
    receiver: AddressType = Field(alias="contract_address")

    @validator("receiver", pre=True, allow_reuse=True)
    def validate_receiver(cls, value):
        if isinstance(value, int):
            return to_checksum_address(value)

        return value

    @validator("max_fee", pre=True, allow_reuse=True)
    def validate_max_fee(cls, value):
        if isinstance(value, str):
            return int(value, 16)

        return value

    @property
    def receiver_int(self) -> int:
        return self.starknet.encode_address(self.receiver)

    @property
    def contract_type(self) -> ContractType:
        return self.chain_manager.contracts[self.receiver]

    @property
    def entry_point_selector(self) -> int:
        return get_selector_from_name(self.method_abi.name)

    @property
    def txn_hash(self) -> HexBytes:
        hash_int = calculate_transaction_hash_common(
            additional_data=[],
            calldata=self.data,
            chain_id=self.provider.chain_id,
            contract_address=self.receiver_int,
            entry_point_selector=self.entry_point_selector,
            max_fee=self.max_fee or 0,
            tx_hash_prefix=TransactionHashPrefix.INVOKE,
            version=self.version,
        )
        return HexBytes(hash_int)

    def as_starknet_object(self) -> InvokeFunction:
        return InvokeFunction(
            calldata=self.data,
            contract_address=self.receiver_int,
            entry_point_selector=self.entry_point_selector,
            signature=[to_int(self.signature.r), to_int(self.signature.s)]
            if self.signature
            else [],
            max_fee=self.max_fee or 0,
            version=self.version,
        )


class StarknetReceipt(ReceiptAPI, StarknetBase):
    """
    An object represented a confirmed transaction in Starknet.
    """

    status: TransactionStatus
    type: TransactionType

    # NOTE: Might be a backend bug causing this to be None
    block_hash: Optional[str] = None
    block_number: Optional[int] = None  # type: ignore

    """Ignored"""
    gas_limit: int = Field(0, exclude=True)
    gas_price: int = Field(0, exclude=True)
    gas_used: int = Field(0, exclude=True)
    sender: str = Field("", exclude=True)

    """Aliased"""
    txn_hash: str = Field(alias="hash")

    @validator("nonce", pre=True, allow_reuse=True)
    def validate(cls, value):
        if isinstance(value, str):
            return int(value, 16)

    @validator("block_hash", pre=True, allow_reuse=True)
    def validate_block_hash(cls, value):
        return HexBytes(value).hex() if value else value

    @validator("txn_hash", pre=True, allow_reuse=True)
    def validate_transaction_hash(cls, value):
        if isinstance(value, int):
            return HexBytes(value).hex()

    @property
    def ran_out_of_gas(self) -> bool:
        return False  # Overidden by Invoke receipts

    @property
    def total_fees_paid(self) -> int:
        return 0  # Overidden by Invoke receipts

    def decode_logs(self, abi: Optional[ContractEventABI] = None) -> Iterator[ContractLog]:
        # Overriden in InvocationReceipt
        pass


class DeployReceipt(StarknetReceipt):
    contract_address: str
    receiver: Optional[str] = None  # type: ignore

    # Only get a receipt if deploy was accepted
    status: TransactionStatus = TransactionStatus.ACCEPTED_ON_L2

    @validator("contract_address", pre=True, allow_reuse=True)
    def validate_contract_address(cls, value):
        if isinstance(value, int):
            return to_checksum_address(value)

        return value


class InvocationReceipt(StarknetReceipt):
    actual_fee: int
    entry_point_selector: Optional[int] = Field(
        default=None, alias="selector"
    )  # Either has this or method_abi
    max_fee: int
    method_abi: Optional[MethodABI] = None  # Either has this or entry_point_selector
    receiver: str = Field(alias="contract_address")
    returndata: List[Any] = Field(default_factory=list, alias="result")
    return_value: List[int] = []

    """Aliased"""
    logs: List[dict] = Field(alias="events")

    @validator("max_fee", pre=True, allow_reuse=True)
    def validate_max_fee(cls, value):
        if isinstance(value, str):
            return int(value, 16)

        return value or 0

    @validator("entry_point_selector", pre=True, allow_reuse=True)
    def validate_entry_point_selector(cls, value):
        if isinstance(value, str):
            return int(value, 16)

        return value

    @validator("logs", pre=True, allow_reuse=True)
    def validate_logs(cls, value):
        if value and isinstance(value[0], Event):
            value = [asdict(event) for event in value]

        return value

    @property
    def ran_out_of_gas(self) -> bool:
        return self.actual_fee >= (self.max_fee or 0)

    @property
    def total_fees_paid(self) -> int:
        return self.actual_fee

    def decode_logs(
        self,
        abi: Optional[ContractEventABI] = None,
    ) -> Iterator[ContractLog]:

        log_data_items: List[Dict] = []
        for log in self.logs:
            log_data = {
                **log,
                "block_hash": self.block_hash,
                "transaction_hash": self.txn_hash,
                "block_number": self.block_number,
            }
            log_data_items.append(log_data)

        if abi is not None:
            if not isinstance(abi, (list, tuple)):
                abi = [abi]

            event_abis: List[EventABI] = [a.abi if not isinstance(a, EventABI) else a for a in abi]
            yield from self.starknet.decode_logs(log_data_items, *event_abis)

        else:
            # If ABI is not provided, decode all events
            addresses = {self.starknet.decode_address(x["from_address"]) for x in log_data_items}
            contract_types = self.chain_manager.contracts.get_multiple(addresses)
            # address → selector → abi
            selectors = {
                address: {get_selector_from_name(e.name): e for e in contract.events}
                for address, contract in contract_types.items()
            }
            for log in log_data_items:
                contract_address = self.starknet.decode_address(log["from_address"])
                if contract_address not in selectors:
                    continue

                for event_key in log.get("keys", []):
                    event_abi = selectors[contract_address][event_key]
                    yield from self.starknet.decode_logs([log], event_abi)


class ContractDeclaration(StarknetReceipt):
    """
    The result of declaring a contract type in Starknet.
    """

    class_hash: int
    receiver: Optional[str] = None  # type: ignore

    # Only get a receipt if deploy was accepted
    status: TransactionStatus = TransactionStatus.ACCEPTED_ON_L2

    type: TransactionType = TransactionType.DECLARE

    @validator("class_hash", pre=True)
    def validate_class_hash(cls, value):
        if isinstance(value, str):
            return int(value, 16)
        elif isinstance(value, bytes):
            return to_int(value)
        else:
            return value

    @cached_property
    def contract_type(
        self,
    ) -> ContractType:
        # Look up contract type by class_hash.
        for contract_name, contract_type in self.project_manager.contracts.items():
            if not contract_type.source_id:
                continue

            program = contract_type.deployment_bytecode
            if not program:
                continue

            code = program.bytecode
            if not code:
                continue

            contract_class = ContractClass.deserialize(HexBytes(code))
            contract_cls = get_contract_class(contract_class=contract_class)
            computed_class_hash = compute_class_hash(contract_cls)
            if computed_class_hash == self.class_hash:
                return contract_type

        raise TransactionError(message="Contract type declaration was not successful.")

    def deploy(self, *args, **kwargs) -> ContractInstance:
        container = ContractContainer(self.contract_type)
        return container.deploy(*args, **kwargs)


__all__ = [
    "ContractDeclaration",
    "DeployTransaction",
    "InvokeFunctionTransaction",
    "StarknetReceipt",
    "StarknetTransaction",
]



import pytest
from ape.api.networks import LOCAL_NETWORK_NAME
from click.testing import CliRunner
from starkware.cairo.lang.vm.cairo_runner import pedersen_hash

from ape_starknet.utils import is_hex_address

from ..conftest import PASSWORD


@pytest.fixture
def isolation():
    return CliRunner().isolation


@pytest.fixture
def devnet_keyfile_account(account_container, account):
    account_container.import_account(
        "__DEV_AS_KEYFILE_ACCOUNT__",
        "testnet",
        account.address,
        private_key=account.private_key,
        passphrase="123",
    )
    return account_container.load("__DEV_AS_KEYFILE_ACCOUNT__")


def test_public_keys(key_file_account, public_key):
    actual = key_file_account.public_key
    assert actual == public_key


def test_sign_message_using_key_file_account(key_file_account, password):
    assert key_file_account.sign_message(5, passphrase=password)


def test_address(account):
    assert is_hex_address(account.address)


def test_sign_message_and_check_signature(account):
    data = 500
    signature = account.sign_message(data)
    data_hash = pedersen_hash(data, 0)
    result = account.check_signature(data_hash, signature)
    assert result, "Failed to validate signature"


def test_sign_message_and_check_signature_using_deployed_account(ephemeral_account):
    data = 500
    signature = ephemeral_account.sign_message(data)
    data_hash = pedersen_hash(data, 0)
    result = ephemeral_account.check_signature(data_hash, signature)
    assert result, "Failed to validate signature"


def test_account_container_contains(account, second_account, key_file_account, account_container):
    assert account.address in account_container
    assert second_account.address in account_container
    assert key_file_account.address in account_container


@pytest.mark.parametrize(
    "get_address",
    [
        lambda a, _: a.address,
        lambda a, e: e.encode_address(a.address),
        lambda a, _: a.public_key,
        lambda a, e: e.encode_address(a.public_key),
    ],
)
def test_access_account_by_str_address(account, account_container, ecosystem, get_address):
    address = get_address(account, ecosystem)
    assert account_container[address] == account
    assert address in account_container


def test_balance(account):
    balance = account.balance
    assert isinstance(balance, int)
    assert account.balance > 0


def test_can_access_devnet_accounts(account, second_account, chain):
    assert chain.contracts[account.address] == account.get_contract_type()
    assert chain.contracts[second_account.address] == second_account.get_contract_type()


def test_import_with_passphrase(account_container, key_file_account):
    alias = "__TEST_IMPORT_WITH_PASSPHRASE__"
    account_container.import_account(
        alias,
        LOCAL_NETWORK_NAME,
        key_file_account.address,
        key_file_account._get_key(PASSWORD),
        passphrase="p@55W0rd",
    )
    new_account = account_container.load(alias)
    assert new_account.address == key_file_account.address


def test_transfer(account, second_account):
    initial_balance = second_account.balance
    account.transfer(second_account, 10)
    assert second_account.balance == initial_balance + 10


def test_unlock_with_passphrase_and_sign_message(isolation, devnet_keyfile_account):
    devnet_keyfile_account.unlock(passphrase="123")

    with isolation(input="y\n"):
        devnet_keyfile_account.sign_message([1, 2, 3])


def test_unlock_from_prompt_and_sign_message(isolation, devnet_keyfile_account):
    with isolation(input="123\n"):
        devnet_keyfile_account.unlock()

    with isolation(input="y\n"):
        devnet_keyfile_account.sign_message([1, 2, 3])


def test_unlock_with_passphrase_and_sign_transaction(isolation, devnet_keyfile_account, contract):
    devnet_keyfile_account.unlock(passphrase="123")

    with isolation(input="y\n"):
        contract.increase_balance(devnet_keyfile_account, 100, sender=devnet_keyfile_account)


def test_unlock_from_prompt_and_sign_transaction(isolation, devnet_keyfile_account, contract):
    with isolation(input="123\n"):
        devnet_keyfile_account.unlock()

    with isolation(input="y\n"):
        contract.increase_balance(devnet_keyfile_account, 100, sender=devnet_keyfile_account)


def test_set_autosign(isolation, devnet_keyfile_account, contract):
    with isolation(input="123\n"):
        devnet_keyfile_account.set_autosign(True)

    contract.increase_balance(devnet_keyfile_account, 100, sender=devnet_keyfile_account)

    # Disable and verify we have to sign again
    devnet_keyfile_account.set_autosign(False)
    with isolation(input="123\ny\n"):
        contract.increase_balance(devnet_keyfile_account, 100, sender=devnet_keyfile_account)


def test_set_autosign_and_provide_passphrase(isolation, devnet_keyfile_account, contract):
    devnet_keyfile_account.set_autosign(True, passphrase="123")
    contract.increase_balance(devnet_keyfile_account, 100, sender=devnet_keyfile_account)

    # Disable and verify we have to sign again
    devnet_keyfile_account.set_autosign(False)
    with isolation(input="123\ny\n"):
        contract.increase_balance(devnet_keyfile_account, 100, sender=devnet_keyfile_account)



from ape import plugins
from ape.api.networks import LOCAL_NETWORK_NAME, NetworkAPI, create_network_type
from ape.types import AddressType

from ape_starknet.accounts import StarknetAccountContracts, StarknetKeyfileAccount
from ape_starknet.config import StarknetConfig
from ape_starknet.conversion import StarknetAddressConverter
from ape_starknet.ecosystems import Starknet
from ape_starknet.explorer import StarknetExplorer
from ape_starknet.provider import StarknetProvider
from ape_starknet.tokens import TokenManager
from ape_starknet.utils import NETWORKS, PLUGIN_NAME

tokens = TokenManager()
network_names = [LOCAL_NETWORK_NAME] + list(NETWORKS.keys())


@plugins.register(plugins.ConversionPlugin)
def converters():
    yield AddressType, StarknetAddressConverter


@plugins.register(plugins.Config)
def config_class():
    return StarknetConfig


@plugins.register(plugins.EcosystemPlugin)
def ecosystems():
    yield Starknet


@plugins.register(plugins.NetworkPlugin)
def networks():
    for network_name, network_params in NETWORKS.items():
        yield PLUGIN_NAME, network_name, create_network_type(*network_params)

    # NOTE: This works for development providers, as they get chain_id from themselves
    yield PLUGIN_NAME, LOCAL_NETWORK_NAME, NetworkAPI


@plugins.register(plugins.ProviderPlugin)
def providers():
    for network_name in network_names:
        yield PLUGIN_NAME, network_name, StarknetProvider


@plugins.register(plugins.AccountPlugin)
def account_types():
    return StarknetAccountContracts, StarknetKeyfileAccount


@plugins.register(plugins.ExplorerPlugin)
def explorers():
    for network_name in network_names:
        yield PLUGIN_NAME, network_name, StarknetExplorer



from typing import TYPE_CHECKING

from ape.api import PluginConfig
from ape.utils import ManagerAccessMixin

if TYPE_CHECKING:
    from ape_starknet.accounts import StarknetAccountContracts
    from ape_starknet.ecosystems import Starknet
    from ape_starknet.explorer import StarknetExplorer
    from ape_starknet.provider import StarknetProvider
    from ape_starknet.tokens import TokenManager


class StarknetBase(ManagerAccessMixin):
    @property
    def starknet_config(self) -> PluginConfig:
        return self.config_manager.get_config("starknet")

    @property
    def starknet(self) -> "Starknet":
        return self.network_manager.starknet  # type: ignore

    @property
    def starknet_explorer(self) -> "StarknetExplorer":
        explorer = self.provider.network.explorer
        assert explorer  # For mypy
        return explorer  # type: ignore

    @property
    def provider(self) -> "StarknetProvider":
        return super().provider  # type: ignore

    @property
    def account_contracts(self) -> "StarknetAccountContracts":
        return self.account_manager.containers["starknet"]  # type: ignore

    @property
    def tokens(self) -> "TokenManager":
        from ape_starknet import tokens

        return tokens



from .conftest import ApeStarknetCliRunner


def test_plugin_loads(ape_cli):
    runner = ApeStarknetCliRunner(ape_cli, ["starknet"])
    output = runner.invoke("--help")
    assert "ERROR" not in runner.invoke("--help"), output



from typing import TYPE_CHECKING, Dict, Union

from ape.api import Address
from ape.contracts import ContractInstance
from ape.exceptions import ContractError
from ape.types import AddressType
from eth_typing import HexAddress, HexStr
from ethpm_types import ContractType
from starknet_devnet.fee_token import FeeToken

from ape_starknet.exceptions import StarknetProviderError
from ape_starknet.utils.basemodel import StarknetBase

if TYPE_CHECKING:
    from ape_starknet.accounts import BaseStarknetAccount


def missing_contract_error(token: str, contract_address: AddressType) -> ContractError:
    return ContractError(f"Incorrect '{token}' contract address '{contract_address}'.")


ERC20 = ContractType(
    **{
        "contractName": "ERC20",
        "abi": [
            {
                "type": "struct",
                "name": "Uint256",
                "members": [
                    {"name": "low", "type": "felt", "offset": 0},
                    {"name": "high", "type": "felt", "offset": 1},
                ],
                "size": 2,
            },
            {
                "type": "event",
                "name": "Transfer",
                "inputs": [
                    {"name": "from_", "type": "felt", "indexed": False},
                    {"name": "to", "type": "felt", "indexed": False},
                    {"name": "value", "type": "Uint256", "indexed": False},
                ],
                "anonymous": False,
            },
            {
                "type": "event",
                "name": "Approval",
                "inputs": [
                    {"name": "owner", "type": "felt", "indexed": False},
                    {"name": "spender", "type": "felt", "indexed": False},
                    {"name": "value", "type": "Uint256", "indexed": False},
                ],
                "anonymous": False,
            },
            {
                "type": "constructor",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "name", "type": "felt"},
                    {"name": "symbol", "type": "felt"},
                    {"name": "decimals", "type": "felt"},
                    {"name": "initial_supply", "type": "Uint256"},
                    {"name": "recipient", "type": "felt"},
                ],
            },
            {
                "type": "function",
                "name": "name",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "name", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "symbol",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "symbol", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "totalSupply",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "totalSupply", "type": "Uint256"}],
            },
            {
                "type": "function",
                "name": "decimals",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "decimals", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "balanceOf",
                "stateMutability": "view",
                "inputs": [{"name": "account", "type": "felt"}],
                "outputs": [{"name": "balance", "type": "Uint256"}],
            },
            {
                "type": "function",
                "name": "allowance",
                "stateMutability": "view",
                "inputs": [{"name": "owner", "type": "felt"}, {"name": "spender", "type": "felt"}],
                "outputs": [{"name": "remaining", "type": "Uint256"}],
            },
            {
                "type": "function",
                "name": "transfer",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "recipient", "type": "felt"},
                    {"name": "amount", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "transferFrom",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "sender", "type": "felt"},
                    {"name": "recipient", "type": "felt"},
                    {"name": "amount", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "approve",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "spender", "type": "felt"},
                    {"name": "amount", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "increaseAllowance",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "spender", "type": "felt"},
                    {"name": "added_value", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "decreaseAllowance",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "spender", "type": "felt"},
                    {"name": "subtracted_value", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
        ],
    }
)


class TokenManager(StarknetBase):
    # The 'test_token' refers to the token that comes with Argent-X
    additional_tokens: Dict = {}
    contract_type = ERC20

    @property
    def token_address_map(self) -> Dict:
        local_eth = self.starknet.decode_address(FeeToken.ADDRESS)
        mainnet_eth = self.starknet.decode_address(
            "0x049d36570d4e46f48e99674bd3fcc84644ddd6b96f7c741b1562b82f9e004dc7"
        )
        testnet_eth = self.starknet.decode_address(
            "0x049d36570d4e46f48e99674bd3fcc84644ddd6b96f7c741b1562b82f9e004dc7"
        )
        testnet_test_token = self.starknet.decode_address(
            "0x07394cbe418daa16e42b87ba67372d4ab4a5df0b05c6e554d158458ce245bc10"
        )
        mainnet_test_token = self.starknet.decode_address(
            "0x07394cbe418daa16e42b87ba67372d4ab4a5df0b05c6e554d158458ce245bc10"
        )

        return {
            "eth": {"local": local_eth, "mainnet": mainnet_eth, "testnet": testnet_eth},
            "test_token": {"testnet": testnet_test_token, "mainnet": mainnet_test_token},
            **self.additional_tokens,
        }

    def __getitem__(self, token: str) -> ContractInstance:
        network = self.provider.network.name
        contract_address = AddressType(
            HexAddress(HexStr(self.token_address_map[token.lower()].get(network)))
        )
        if not contract_address:
            raise IndexError(f"No token '{token}'.")

        return ContractInstance(contract_address, ERC20)

    def is_token(self, address: AddressType) -> bool:
        network = self.provider.network.name
        return any(address == networks.get(network) for networks in self.token_address_map.values())

    def add_token(self, name: str, network: str, address: AddressType):
        if name not in self.additional_tokens:
            self.additional_tokens[name] = {}

        self.additional_tokens[name][network] = address

    def get_balance(self, account: Union[Address, AddressType], token: str = "eth") -> int:
        if hasattr(account, "address"):
            account = account.address  # type: ignore

        return self[token].balanceOf(account)

    def transfer(
        self,
        sender: Union[int, AddressType, "BaseStarknetAccount"],
        receiver: Union[int, AddressType, "BaseStarknetAccount"],
        amount: int,
        token: str = "eth",
    ):
        if isinstance(receiver, int):
            receiver_address = receiver
        elif hasattr(receiver, "address_int"):
            receiver_address = receiver.address_int  # type: ignore
        elif isinstance(receiver, str):
            receiver_address = self.starknet.encode_address(receiver)
        else:
            raise StarknetProviderError(
                f"Unhandled type for receiver '{receiver}'. Expects int, str, or account."
            )

        sender_account = (
            self.account_contracts[sender] if isinstance(sender, (int, str)) else sender
        )
        return self[token].transfer(receiver_address, amount, sender=sender_account)



from ape.contracts import ContractInstance


def test_multiple_networks_in_test(
    networks, provider, eth_account, eth_contract_container, contract, account
):
    starknet_chain_id = provider.chain_id

    with networks.ethereum.local.use_provider("test") as eth_provider:
        # Verify the chain changed
        assert eth_provider.chain_id != starknet_chain_id

        # Deploy and interact with a contract on Ethereum
        eth_contract = eth_contract_container.deploy(sender=eth_account)
        receipt = eth_contract.setNumber(123, sender=eth_account)
        assert not receipt.failed
        assert isinstance(eth_contract, ContractInstance)
        assert eth_contract.myNumber() == 123

    # Switch back to Starknet
    receipt = contract.increase_balance(account.address, 123, sender=account)
    assert not receipt.failed



import functools
import json
import random
from dataclasses import dataclass
from math import ceil
from pathlib import Path
from typing import Dict, Iterable, Iterator, List, Optional, Union

import click
from ape.api import AccountAPI, AccountContainerAPI, ReceiptAPI, TransactionAPI
from ape.api.address import BaseAddress
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractContainer
from ape.exceptions import AccountsError, SignatureError
from ape.logging import LogLevel, logger
from ape.types import AddressType, SignableMessage, TransactionSignature
from ape.utils import abstractmethod, cached_property
from eth_keyfile import create_keyfile_json, decode_keyfile_json
from eth_typing import HexAddress, HexStr
from eth_utils import add_0x_prefix, text_if_str, to_bytes
from ethpm_types import ContractType
from ethpm_types.abi import MethodABI
from hexbytes import HexBytes
from starknet_py.net import KeyPair
from starknet_py.net.account.compiled_account_contract import COMPILED_ACCOUNT_CONTRACT
from starknet_py.net.signer.stark_curve_signer import StarkCurveSigner
from starknet_py.utils.crypto.facade import ECSignature, message_signature, pedersen_hash
from starkware.cairo.lang.vm.cairo_runner import verify_ecdsa_sig
from starkware.crypto.signature.signature import private_to_stark_key
from starkware.starknet.core.os.contract_address.contract_address import (
    calculate_contract_address_from_hash,
)
from starkware.starknet.services.api.contract_class import ContractClass

from ape_starknet.exceptions import StarknetProviderError
from ape_starknet.tokens import TokenManager
from ape_starknet.transactions import InvokeFunctionTransaction
from ape_starknet.utils import (
    convert_contract_class_to_contract_type,
    get_chain_id,
    get_random_private_key,
    pad_hex_str,
)
from ape_starknet.utils.basemodel import StarknetBase

APP_KEY_FILE_KEY = "ape-starknet"
"""
The key-file stanza containing custom properties
specific to the ape-starknet plugin.
"""
APP_KEY_FILE_VERSION = "0.1.0"
OZ_CONTRACT_CLASS = ContractClass.loads(COMPILED_ACCOUNT_CONTRACT)
OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE = convert_contract_class_to_contract_type(OZ_CONTRACT_CLASS)

# https://github.com/starkware-libs/cairo-lang/blob/v0.9.1/src/starkware/starknet/cli/starknet_cli.py#L66
FEE_MARGIN_OF_ESTIMATION = 1.1


def sign_calldata(calldata: Iterable[int], priv_key: int):
    """
    Helper function that signs hash:

        hash = pedersen_hash(calldata[0], 0)
        hash = pedersen_hash(calldata[1], hash)
        hash = pedersen_hash(calldata[2], hash)
        ...

    :param calldata: iterable of ints
    :param priv_key: private key
    :return: signed calldata's hash
    """
    hashed_calldata = functools.reduce(lambda x, y: pedersen_hash(y, x), calldata, 0)
    return message_signature(hashed_calldata, priv_key)


class StarknetAccountContracts(AccountContainerAPI, StarknetBase):

    ephemeral_accounts: Dict[str, Dict] = {}
    """Local-network accounts that do not persist."""

    cached_accounts: Dict[str, "StarknetKeyfileAccount"] = {}
    """Accounts created in a live network that persist in key-files."""

    @property
    def provider_config(self) -> Dict:
        return self.starknet_config["provider"]

    @property
    def number_of_devnet_accounts(self) -> int:
        if not self.network_manager.active_provider:
            return 0

        if self.provider.network.name != LOCAL_NETWORK_NAME:
            return 0

        return self.provider_config.local["number_of_accounts"]  # type: ignore

    @property
    def devnet_account_seed(self) -> int:
        return self.provider_config.local["seed"]  # type: ignore

    @property
    def _key_file_paths(self) -> Iterator[Path]:
        for path in self.data_folder.glob("*.json"):
            if path.stem not in ("deployments_map",):
                yield path

    @property
    def aliases(self) -> Iterator[str]:
        yield from self.ephemeral_accounts.keys()
        for key_file in self._key_file_paths:
            yield key_file.stem

    @property
    def public_key_addresses(self) -> Iterator[AddressType]:
        for account in self.accounts:
            yield account.address

    @cached_property
    def test_accounts(self) -> List["StarknetDevnetAccount"]:
        random_generator = random.Random()
        random_generator.seed(self.devnet_account_seed)
        devnet_accounts = [
            StarknetDevnetAccount(private_key=random_generator.getrandbits(128))
            for _ in range(self.number_of_devnet_accounts)
        ]

        # Track all devnet account contracts in chain manager for look-up purposes
        for account in devnet_accounts:
            self.chain_manager.contracts[account.address] = account.get_contract_type()

        return devnet_accounts

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for test_account in self.test_accounts:
            yield test_account

        for alias, account_data in self.ephemeral_accounts.items():
            yield StarknetEphemeralAccount(raw_account_data=account_data, account_key=alias)

        for key_file_path in self._key_file_paths:
            if key_file_path.stem == "deployments_map":
                continue

            if key_file_path.stem in self.cached_accounts:
                yield self.cached_accounts[key_file_path.stem]
            else:
                account = StarknetKeyfileAccount(key_file_path=key_file_path)
                self.cached_accounts[key_file_path.stem] = account
                yield account

    def __len__(self) -> int:
        return len([*self._key_file_paths])

    def __setitem__(self, address: AddressType, account: AccountAPI):
        pass

    def __delitem__(self, address: AddressType):
        pass

    def __getitem__(self, item: Union[AddressType, int]) -> AccountAPI:
        address = HexAddress(HexStr(HexBytes(item).hex())) if isinstance(item, int) else item

        # First, check if user accessing via public key
        for account in self.accounts:
            if not isinstance(account, BaseStarknetAccount):
                continue

            if account.public_key == address:
                return super().__getitem__(account.address)

        # Else, use the contract address (more expected)
        checksum_address = self.starknet.decode_address(address)
        return super().__getitem__(checksum_address)

    def get_account(self, address: Union[AddressType, int]) -> "BaseStarknetAccount":
        return self[address]  # type: ignore

    def load(self, alias: str) -> "BaseStarknetAccount":
        if alias in self.ephemeral_accounts:
            return StarknetEphemeralAccount(
                raw_account_data=self.ephemeral_accounts[alias], account_key=alias
            )

        return self.load_key_file_account(alias)

    def load_key_file_account(self, alias: str) -> "StarknetKeyfileAccount":
        if alias in self.cached_accounts:
            return self.cached_accounts[alias]

        for key_file_path in self._key_file_paths:
            if key_file_path.stem == alias:
                account = StarknetKeyfileAccount(key_file_path=key_file_path)
                self.cached_accounts[alias] = account
                return account

        raise AccountsError(f"Starknet account '{alias}' not found.")

    def import_account_from_key_file(self, alias: str, key_file: Path):
        if not key_file.is_file():
            raise AccountsError(f"Unknown keyfile '{key_file}'.")

        destination = self.data_folder.joinpath(f"{alias}.json")
        if destination.exists():
            raise AccountsError(f"Account already saved with alias '{alias}'.")

        key_file_data = json.loads(key_file.read_text())
        if "argent" in key_file_data and APP_KEY_FILE_KEY not in key_file_data:
            # Migrate Argent-X keyfile

            deployments = []
            for account in key_file_data["argent"]["accounts"]:
                network = _clean_network_name(account["network"])
                deployment = StarknetAccountDeployment(
                    network_name=network, contract_address=account["address"]
                )
                deployments.append(vars(deployment))

            key_file_data = {**key_file_data, **_create_key_file_app_data(deployments)}

        destination.write_text(json.dumps(key_file_data))

    def import_account(
        self,
        alias: str,
        network_name: str,
        contract_address: str,
        private_key: Union[int, str],
        passphrase: Optional[str] = None,
    ):
        address = self.starknet.decode_address(contract_address)
        if isinstance(private_key, str) and private_key.startswith("0x"):
            private_key = pad_hex_str(private_key.strip("'\""))
            private_key = int(private_key, 16)
        elif isinstance(private_key, str):
            private_key = int(private_key)

        network_name = _clean_network_name(network_name)
        key_pair = KeyPair.from_private_key(private_key)
        deployments = [{"network_name": network_name, "contract_address": address}]

        if network_name == LOCAL_NETWORK_NAME:
            account_data = {
                "address": key_pair.public_key,
                "private_key": key_pair.private_key,
                **_create_key_file_app_data(deployments),
            }
            self.ephemeral_accounts[alias] = account_data
        else:
            # Only write keyfile if not in a local network
            path = self.data_folder.joinpath(f"{alias}.json")
            new_account = StarknetKeyfileAccount(key_file_path=path)
            new_account.write(
                passphrase=passphrase, private_key=private_key, deployments=deployments
            )

        # Ensure contract gets cached
        network = self.starknet.get_network(network_name)
        with network.use_provider(network.default_provider or "starknet"):
            contract_type = self.starknet_explorer.get_contract_type(address)
            if not contract_type:
                raise StarknetProviderError(f"Failed to get contract type for account '{address}'.")

            self.chain_manager.contracts[address] = contract_type

    def deploy_account(
        self, alias: str, private_key: Optional[int] = None, token: Optional[str] = None
    ) -> str:
        """
        Deploys an account contract for the given alias.

        Args:
            alias (str): The alias to use to reference the account in ``ape``.
            private_key (Optional[int]): Optionally provide your own private key.`
            token (Optional[str]): Used for deploying contracts in Alpha MainNet.

        Returns:
            str: The contract address of the account.
        """

        if alias in self.aliases:
            raise AccountsError(f"Account with alias '{alias}' already exists.")

        network_name = self.provider.network.name
        logger.info(f"Deploying an account to '{network_name}' network ...")

        private_key = private_key or int(get_random_private_key(), 16)
        key_pair = KeyPair.from_private_key(private_key)

        account_container = ContractContainer(contract_type=OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE)
        instance = account_container.deploy(key_pair.public_key, token=token)
        self.import_account(alias, network_name, instance.address, key_pair.private_key)
        return instance.address

    def delete_account(
        self, alias: str, network: Optional[str] = None, passphrase: Optional[str] = None
    ):
        network = _clean_network_name(network) if network else self.provider.network.name
        if alias in self.ephemeral_accounts:
            del self.ephemeral_accounts[alias]
        else:
            account = self.load_key_file_account(alias)
            account.delete(network, passphrase=passphrase)


@dataclass
class StarknetAccountDeployment:
    network_name: str
    contract_address: AddressType


class BaseStarknetAccount(AccountAPI, StarknetBase):
    token_manager: TokenManager = TokenManager()

    @abstractmethod
    def _get_key(self) -> int:
        ...

    @abstractmethod
    def get_account_data(self) -> Dict:
        ...

    @abstractmethod
    def get_contract_type(self) -> ContractType:
        ...

    @property
    def address(self) -> AddressType:
        for deployment in self.get_deployments():
            network_name = deployment.network_name
            network = self.starknet.networks[network_name]
            if network_name == network.name:
                contract_address = deployment.contract_address
                return self.starknet.decode_address(contract_address)

        raise AccountsError("Account not deployed.")

    @property
    def address_int(self) -> int:
        return self.starknet.encode_address(self.address)

    @property
    def public_key(self) -> str:
        account_data = self.get_account_data()
        if "address" not in account_data:
            raise StarknetProviderError(
                f"Account data corrupted, missing 'address' key: {account_data}."
            )

        address = account_data["address"]
        if isinstance(address, int):
            address = HexBytes(address).hex()

        return add_0x_prefix(address)

    @cached_property
    def signer(self) -> StarkCurveSigner:
        key_pair = KeyPair.from_private_key(self._get_key())
        return StarkCurveSigner(
            account_address=self.address,
            key_pair=key_pair,
            chain_id=get_chain_id(self.provider.chain_id),
        )

    @cached_property
    def execute_abi(self) -> Optional[MethodABI]:
        contract_type = self.get_contract_type()
        if not contract_type:
            return None

        execute_abi_ls = [
            abi for abi in contract_type.abi if getattr(abi, "name", "") == "__execute__"
        ]
        if not execute_abi_ls:
            raise AccountsError(f"Account '{self.address}' does not have __execute__ method.")

        abi = execute_abi_ls[0]
        if not isinstance(abi, MethodABI):
            raise AccountsError("ABI for '__execute__' is not a method.")

        return abi

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.address}>"

    def call(self, txn: TransactionAPI, send_everything: bool = False) -> ReceiptAPI:
        if send_everything:
            raise NotImplementedError("send_everything currently isn't implemented in Starknet.")

        if not isinstance(txn, InvokeFunctionTransaction):
            raise AccountsError("Can only call Starknet transactions.")

        txn = self.prepare_transaction(txn)
        if not txn.signature:
            raise SignatureError("The transaction was not signed.")

        return self.provider.send_transaction(txn)

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        self._prepare_transaction(txn)
        if txn.max_fee is None:
            # NOTE: Signature cannot be None when estimating fees.
            txn.signature = self.sign_transaction(txn)
            txn.max_fee = ceil(self.get_fee_estimate(txn) * FEE_MARGIN_OF_ESTIMATION)

        txn.signature = self.sign_transaction(txn)
        return txn

    def get_fee_estimate(self, txn: TransactionAPI) -> int:
        return self.provider.estimate_gas_cost(txn)

    def _prepare_transaction(self, txn: TransactionAPI):
        execute_abi = self.execute_abi
        if not execute_abi:
            raise AccountsError(
                f"Account is not deployed to network '{self.provider.network.name}'."
            )

        if not isinstance(txn, InvokeFunctionTransaction):
            raise AccountsError("Can only prepare invoke transactions.")

        txn: InvokeFunctionTransaction = super().prepare_transaction(txn)  # type: ignore
        stark_tx = txn.as_starknet_object()
        account_call = {
            "to": stark_tx.contract_address,
            "selector": stark_tx.entry_point_selector,
            "data_offset": 0,
            "data_len": len(stark_tx.calldata),
        }
        contract_type = self.chain_manager.contracts[self.address]
        txn.data = self.starknet.encode_calldata(
            contract_type.abi, execute_abi, [[account_call], stark_tx.calldata, self.nonce]
        )
        txn.receiver = self.address
        txn.sender = None
        txn.original_method_abi = txn.method_abi
        txn.method_abi = execute_abi

    def sign_transaction(self, txn: TransactionAPI) -> TransactionSignature:
        if not isinstance(txn, InvokeFunctionTransaction):
            raise AccountsError("This account can only sign Starknet transactions.")

        # NOTE: 'v' is not used
        sign_result = self.signer.sign_transaction(txn.as_starknet_object())
        if not sign_result:
            raise SignatureError("Failed to sign transaction.")

        r = to_bytes(sign_result[0])
        s = to_bytes(sign_result[1])
        return TransactionSignature(v=0, r=r, s=s)  # type: ignore

    def transfer(
        self,
        account: Union[str, AddressType, BaseAddress],
        value: Union[str, int, None] = None,
        data: Union[bytes, str, None] = None,
        **kwargs,
    ) -> ReceiptAPI:
        value = value or 0
        value = self.conversion_manager.convert(value, int) or 0
        if not isinstance(value, int):
            if value.isnumeric():
                value = str(value)
            else:
                raise StarknetProviderError("value is not an integer.")

        if not isinstance(account, str) and hasattr(account, "address"):
            receiver = getattr(account, "address")

        elif isinstance(account, str):
            checksummed_address = self.starknet.decode_address(account)
            receiver = self.starknet.encode_address(checksummed_address)

        elif isinstance(account, int):
            receiver = account

        else:
            raise TypeError(f"Unable to handle account type '{type(account)}'.")

        return self.token_manager.transfer(self.address, receiver, value, **kwargs)

    def get_deployment(self, network_name: str) -> Optional[StarknetAccountDeployment]:
        return next(
            (
                deployment
                for deployment in self.get_deployments()
                if deployment.network_name in network_name
            ),
            None,
        )

    def check_signature(  # type: ignore
        self,
        data: int,
        signature: Optional[ECSignature] = None,  # TransactionAPI doesn't need it
    ) -> bool:
        public_key_int = self.starknet.encode_address(self.public_key)
        return verify_ecdsa_sig(public_key_int, data, signature)

    def get_deployments(self) -> List[StarknetAccountDeployment]:
        plugin_key_file_data = self.get_account_data().get(APP_KEY_FILE_KEY, {})
        return [StarknetAccountDeployment(**d) for d in plugin_key_file_data.get("deployments", [])]


class StarknetDevelopmentAccount(BaseStarknetAccount):
    def get_contract_type(self) -> ContractType:
        return OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE

    def sign_message(self, msg: SignableMessage) -> Optional[ECSignature]:
        if not isinstance(msg, (list, tuple)):
            msg = [msg]

        return sign_calldata(msg, self._get_key())  # type: ignore


class StarknetDevnetAccount(StarknetDevelopmentAccount):
    """
    Accounts generated in the starknet-devnet process.
    """

    private_key: int

    @cached_property
    def public_key_int(self) -> int:
        return private_to_stark_key(self.private_key)

    @cached_property
    def public_key(self) -> str:
        return add_0x_prefix(HexStr(HexBytes(self.public_key_int).hex()))

    @cached_property
    def address(self) -> AddressType:
        address_int = calculate_contract_address_from_hash(
            # Hardcoded values since devnet 0.2.6:
            # https://github.com/Shard-Labs/starknet-devnet/blob/v0.2.6/starknet_devnet/account.py#L36
            salt=20,
            class_hash=1803505466663265559571280894381905521939782500874858933595227108099796801620,
            constructor_calldata=[self.public_key_int],
            deployer_address=0,
        )
        return self.starknet.decode_address(address_int)

    def _get_key(self) -> int:
        return self.private_key

    def get_account_data(self) -> Dict:
        deployments = [
            {
                "contract_address": self.address,
                "network_name": LOCAL_NETWORK_NAME,
            }
        ]
        return {
            "private_key": self.private_key,
            "public_key": self.public_key,
            APP_KEY_FILE_KEY: {"deployments": deployments},
        }


class StarknetEphemeralAccount(StarknetDevelopmentAccount):
    """
    Accounts deployed on a local Starknet chain.
    """

    raw_account_data: Dict
    account_key: str

    def get_account_data(self) -> Dict:
        return self.raw_account_data

    @property
    def alias(self) -> Optional[str]:
        return self.account_key

    def _get_key(self) -> int:
        if "private_key" not in self.raw_account_data:
            raise AccountsError("This account cannot sign.")

        return self.raw_account_data["private_key"]


class StarknetKeyfileAccount(BaseStarknetAccount):
    key_file_path: Path
    locked: bool = True
    __autosign: bool = False
    __cached_key: Optional[int] = None

    @property
    def alias(self) -> Optional[str]:
        return self.key_file_path.stem

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        self._prepare_transaction(txn)
        do_relock = False
        if not txn.max_fee:
            if self.locked:
                # Unlock to prevent multiple prompts for signing transaction.
                original_level = logger.level
                logger.set_level(LogLevel.ERROR)
                self.set_autosign(True)
                logger.set_level(original_level)

            txn.signature = self.sign_transaction(txn)
            txn.max_fee = ceil(self.get_fee_estimate(txn) * FEE_MARGIN_OF_ESTIMATION)

        txn.signature = self.sign_transaction(txn)

        if do_relock:
            self.locked = True
            self.set_autosign(False)

        return txn

    def get_contract_type(self) -> ContractType:
        return OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE

    def write(
        self,
        passphrase: Optional[str] = None,
        private_key: Optional[int] = None,
        deployments: Optional[List[Dict]] = None,
    ):
        passphrase = (
            click.prompt("Enter a new passphrase", hide_input=True, confirmation_prompt=True)
            if passphrase is None
            else passphrase
        )
        key_file_data = self.__encrypt_key_file(passphrase, private_key=private_key)
        account_data = self.get_account_data()
        if deployments:
            if APP_KEY_FILE_KEY not in account_data:
                account_data[APP_KEY_FILE_KEY] = {}

            account_data[APP_KEY_FILE_KEY]["deployments"] = deployments

        account_data = {**account_data, **key_file_data}
        self.key_file_path.write_text(json.dumps(account_data))

    def get_account_data(self) -> Dict:
        if self.key_file_path.is_file():
            return json.loads(self.key_file_path.read_text())

        return {}

    def delete(self, network: str, passphrase: Optional[str] = None):
        passphrase = (
            click.prompt(
                f"Enter passphrase to delete '{self.alias}'",
                hide_input=True,
            )
            if passphrase is None
            else passphrase
        )

        try:
            self.__decrypt_key_file(passphrase)
        except FileNotFoundError:
            return

        network = _clean_network_name(network)
        deployments = self.get_deployments()
        if network not in [d.network_name for d in deployments]:
            raise AccountsError(f"Account '{self.alias}' not deployed to network '{network}'.")

        remaining_deployments = [
            vars(d) for d in self.get_deployments() if d.network_name != network
        ]
        if remaining_deployments:
            self.write(passphrase=passphrase, deployments=remaining_deployments)
        elif click.confirm(f"Completely delete local key for account '{self.address}'?"):
            # Delete entire account JSON if no more deployments.
            # The user has to agree to an additional prompt since this may be very destructive.
            self.key_file_path.unlink()

    def sign_message(
        self, msg: SignableMessage, passphrase: Optional[str] = None
    ) -> Optional[ECSignature]:
        if not isinstance(msg, (list, tuple)):
            msg = [msg]

        private_key = self._get_key(passphrase=passphrase)
        return sign_calldata(msg, private_key)  # type: ignore

    def change_password(self):
        self.locked = True  # force entering passphrase to get key
        original_passphrase = self._get_passphrase_from_prompt()
        private_key = self._get_key(passphrase=original_passphrase)
        self.write(passphrase=None, private_key=private_key)

    def add_deployment(self, network_name: str, contract_address: AddressType):
        passphrase = self._get_passphrase_from_prompt()
        network_name = _clean_network_name(network_name)
        deployments = [
            vars(d) for d in self.get_deployments() if d.network_name not in network_name
        ]
        new_deployment = StarknetAccountDeployment(
            network_name=network_name, contract_address=contract_address
        )
        deployments.append(vars(new_deployment))

        self.write(
            passphrase=passphrase,
            private_key=self._get_key(passphrase=passphrase),
            deployments=deployments,
        )

    def unlock(self, passphrase: Optional[str] = None):
        passphrase = passphrase or self._get_passphrase_from_prompt(
            f"Enter passphrase to unlock '{self.alias}'"
        )
        self._get_key(passphrase=passphrase)
        self.locked = False

    def set_autosign(self, enabled: bool, passphrase: Optional[str] = None):
        if enabled:
            self.unlock(passphrase=passphrase)
            logger.warning("Danger! This account will now sign any transaction its given.")

        self.__autosign = enabled
        if not enabled:
            # Re-lock if was turning off
            self.locked = True
            self.__cached_key = None

    def _get_key(self, passphrase: Optional[str] = None) -> int:
        if self.__cached_key is not None:
            if not self.locked:
                click.echo(f"Using cached key for '{self.alias}'")
                return self.__cached_key
            else:
                self.__cached_key = None

        if passphrase is None:
            passphrase = self._get_passphrase_from_prompt()

        key_hex_str = self.__decrypt_key_file(passphrase).hex()
        key = int(key_hex_str, 16)
        if self.locked and (
            passphrase is not None or click.confirm(f"Leave '{self.alias}' unlocked?")
        ):
            self.locked = False
            self.__cached_key = key

        return key

    def _get_passphrase_from_prompt(self, message: Optional[str] = None) -> str:
        message = message or f"Enter passphrase to unlock '{self.alias}'"
        return click.prompt(
            message,
            hide_input=True,
            default="",  # Just in case there's no passphrase
        )

    def __encrypt_key_file(self, passphrase: str, private_key: Optional[int] = None) -> Dict:
        private_key = self._get_key(passphrase=passphrase) if private_key is None else private_key
        key_str = pad_hex_str(HexBytes(private_key).hex())
        passphrase_bytes = text_if_str(to_bytes, passphrase)
        return create_keyfile_json(HexBytes(key_str), passphrase_bytes, kdf="scrypt")

    def __decrypt_key_file(self, passphrase: str) -> HexBytes:
        key_file_dict = json.loads(self.key_file_path.read_text())
        password_bytes = text_if_str(to_bytes, passphrase)
        decoded_json = decode_keyfile_json(key_file_dict, password_bytes)
        return HexBytes(decoded_json)


def _clean_network_name(network: str) -> str:
    for net in ("local", "mainnet", "testnet"):
        if net in network:
            return net

    if "goerli" in network:
        return "testnet"

    return network


def _create_key_file_app_data(deployments: List[Dict[str, str]]) -> Dict:
    return {APP_KEY_FILE_KEY: {"version": APP_KEY_FILE_VERSION, "deployments": deployments}}


__all__ = [
    "StarknetAccountContracts",
    "StarknetEphemeralAccount",
    "StarknetKeyfileAccount",
]



from typing import List, Optional, Union

import pytest
from ape._cli import cli
from click.testing import CliRunner


@pytest.fixture(scope="session")
def ape_cli():
    yield cli


class ApeStarknetCliRunner:
    runner = CliRunner()

    def __init__(self, cli, base_cmd: List[str]):
        self._cli = cli
        self.base_cmd = base_cmd

    def invoke(
        self, *cmd, input: Optional[Union[str, List[str]]] = None, ensure_successful: bool = True
    ):
        if isinstance(input, (list, tuple)):
            input_str = "\n".join([str(i) for i in input])
        else:
            input_str = input or ""

        ape_cmd = self._get_cmd(*cmd)
        catch_exceptions = not ensure_successful
        result = self.runner.invoke(
            self._cli, ape_cmd, catch_exceptions=catch_exceptions, input=input_str
        )

        if ensure_successful:
            assert result.exit_code == 0, result.output

        return result.output

    def _get_cmd(self, *args) -> List[str]:
        return [*self.base_cmd, *args]



import pytest

from ape_starknet.exceptions import StarknetProviderError


@pytest.fixture(scope="module")
def testnet_provider(networks):
    with networks.parse_network_choice("starknet:testnet") as provider:
        yield provider


@pytest.fixture(scope="module")
def eth_contract(testnet_provider):
    return testnet_provider.tokens["eth"]


def test_can_connect_to_testnet(eth_contract):
    eth_contract.balanceOf("0x348ef2b95e31269b4a1c019428723e3a33cd964f92ad866741f189b88be3bc0")


def test_is_token(eth_contract, tokens):
    # Ensure it's recognized as a token
    assert tokens.is_token(eth_contract)


def test_revert_message_no_account_found(eth_contract, account):
    # It will obviously fail because we are using a local account
    reason = "No contract found for identifier.*"
    with pytest.raises(StarknetProviderError, match=reason):
        eth_contract.increaseAllowance(account, 1, sender=account)



from ape.api import PluginConfig
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.utils import DEFAULT_NUMBER_OF_TEST_ACCOUNTS

from ape_starknet.utils import DEFAULT_ACCOUNT_SEED

DEFAULT_PORT = 8545


class NetworkConfig(PluginConfig):
    required_confirmations: int = 0
    block_time: int = 0
    default_provider: str = "starknet"


class ProviderConfig(PluginConfig):
    mainnet: dict = {"uri": "https://alpha-mainnet.starknet.io"}
    testnet: dict = {"uri": "https://alpha4.starknet.io"}
    local: dict = {
        "uri": f"http://127.0.0.1:{DEFAULT_PORT}",
        "seed": DEFAULT_ACCOUNT_SEED,
        "number_of_accounts": DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
    }


class StarknetConfig(PluginConfig):
    mainnet: NetworkConfig = NetworkConfig(required_confirmations=7, block_time=13)
    testnet: NetworkConfig = NetworkConfig(required_confirmations=2, block_time=15)
    local: NetworkConfig = NetworkConfig()
    default_network: str = LOCAL_NETWORK_NAME
    provider: ProviderConfig = ProviderConfig()



import json
import random
import tempfile
from pathlib import Path

import pytest

from ape_starknet.utils import get_random_private_key

from ..conftest import CONTRACT_ADDRESS, EXISTING_KEY_FILE_ALIAS, PASSWORD
from .conftest import ApeStarknetCliRunner


@pytest.fixture(scope="module")
def accounts_runner(ape_cli):
    return ApeStarknetCliRunner(ape_cli, ["starknet", "accounts"])


@pytest.fixture(scope="module")
def root_accounts_runner(ape_cli):
    return ApeStarknetCliRunner(ape_cli, ["accounts"])


@pytest.fixture(scope="module")
def argent_x_backup(argent_x_key_file_account_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        key_file_path = Path(temp_dir) / "argent-x-backup.json"
        key_file_path.write_text(json.dumps(argent_x_key_file_account_data))
        yield key_file_path


def test_create(accounts_runner):
    """
    Integration test for accounts.
    It all happens under one test because it is really slow to deploy accounts.
    """

    random_alias = "".join(random.choice(["a", "b", "c", "d", "e", "f"]) for _ in range(6))
    output = accounts_runner.invoke("create", random_alias)
    assert "Account successfully deployed to" in output


def test_delete(accounts_runner, key_file_account):
    """
    This integration test deletes a single deployment of an account and then
    re-import it. The account never completely goes away because it is deployed on
    multiple networks.
    """
    output = accounts_runner.invoke(
        "delete",
        EXISTING_KEY_FILE_ALIAS,
        "--network",
        "starknet:testnet",
        input=PASSWORD,
    )
    assert EXISTING_KEY_FILE_ALIAS in output
    output = accounts_runner.invoke("list")

    # The account should still have a remaining deployment and thus still found in output
    assert EXISTING_KEY_FILE_ALIAS in output

    # Only the testnet deployment should have gotten deleted
    output_lines = output.split("\n")
    for index in range(len(output_lines)):
        line = output_lines[index]
        if EXISTING_KEY_FILE_ALIAS not in line:
            continue

        # Find the deployments parts
        next_index = index + 1
        for next_line in output_lines[next_index:]:
            if not next_line.strip():
                break  # Reached end of account's section

            if next_line.startswith("Contract address (testnet)"):
                assert CONTRACT_ADDRESS not in next_line, "Deployment failed to delete"

    # Re-import the deployment (clean-up)
    accounts_runner.invoke(
        "import",
        EXISTING_KEY_FILE_ALIAS,
        "--network",
        "starknet:testnet",
        "--address",
        CONTRACT_ADDRESS,
        input=PASSWORD,
    )


@pytest.mark.parametrize(
    "private_key",
    (
        get_random_private_key(),
        "0x0097a6a4998e2eb47d4cea623c9f8b3048764fc38a92616bdf1c3e68be8b5e28",
        267944034277627769235577208827196223019601239705086925741947749358138777128,
    ),
)
def test_import(accounts_runner, key_file_account, account_container, private_key):
    network = "starknet:testnet"  # NOTE: Does not actually connect
    account_path = account_container.data_folder / f"{EXISTING_KEY_FILE_ALIAS}.json"
    address = key_file_account.address

    if account_path.is_file():
        # Corrupted from previous test
        account_path.unlink()

    accounts_runner.invoke(
        "import",
        EXISTING_KEY_FILE_ALIAS,
        "--network",
        network,
        "--address",
        address,
        input=[private_key, PASSWORD, PASSWORD],
    )

    # Clean-up
    accounts_runner.invoke(
        "delete",
        EXISTING_KEY_FILE_ALIAS,
        "--network",
        network,
        input=[PASSWORD, "y"],
    )


def test_import_argent_x_key_file(accounts_runner, argent_x_backup, account_container):
    alias = "__TEST_ARGENT_X_BACKUP__"
    account_path = account_container.data_folder / f"{alias}.json"

    if account_path.is_file():
        # Corrupted from previous test
        account_path.unlink()

    output = accounts_runner.invoke(
        "import",
        alias,
        "--keyfile",
        str(argent_x_backup),
        "--network",
        "starknet:testnet",
        input=PASSWORD,
    )
    assert "SUCCESS" in output
    account_path.unlink()


def test_import_when_local(accounts_runner):
    output = accounts_runner.invoke(
        "import",
        "FAILS",
        "--address",
        "0x0098580e36aB1485C66f0DC95C2c923e734B7Af44D04dD2B5b9d0809Aa672033",
        ensure_successful=False,
    )
    assert "ERROR: Must use --network option to specify non-local network." in output


def test_list(accounts_runner, key_file_account):
    assert EXISTING_KEY_FILE_ALIAS in accounts_runner.invoke("list")


def test_core_accounts_list_all(root_accounts_runner, key_file_account):
    # This is making sure the normal `ape accounts list --all` command works.
    assert EXISTING_KEY_FILE_ALIAS in root_accounts_runner.invoke("list", "--all")


def test_change_password(accounts_runner, key_file_account):
    new_password = "321"
    assert "SUCCESS" in accounts_runner.invoke(
        "change-password", EXISTING_KEY_FILE_ALIAS, input=[PASSWORD, new_password, new_password]
    )
    assert "SUCCESS" in accounts_runner.invoke(
        "change-password", EXISTING_KEY_FILE_ALIAS, input=[new_password, PASSWORD, PASSWORD]
    )
